import time
import sympy
import platform

import signal
# https://stackoverflow.com/a/22348885/538379
if platform.system() == 'Windows':
    class timeout:
        def __init__(self, seconds=1, error_message='Timeout'):
            self.seconds = seconds
            self.error_message = error_message
        def handle_timeout(self, signum, frame):
            raise TimeoutError(self.error_message)
        def __enter__(self):
            pass
        def __exit__(self, type, value, traceback):
            pass
else:
    class timeout:
        def __init__(self, seconds=1, error_message='Timeout'):
            self.seconds = seconds
            self.error_message = error_message
        def handle_timeout(self, signum, frame):
            raise TimeoutError(self.error_message)
        def __enter__(self):
            signal.signal(signal.SIGALRM, self.handle_timeout)
            signal.alarm(self.seconds)
        def __exit__(self, type, value, traceback):
            signal.alarm(0)




# Source: http://stackoverflow.com/questions/5849800/tic-toc-functions-analog-in-python
_tstart_stack = []

def tic():
    _tstart_stack.append(time.time())

def toc(show=False,fmt="Elapsed: %s s"):
    if show:
        print(fmt % (time.time() - _tstart_stack.pop()))
    else:
        return (time.time() - _tstart_stack.pop())

class Timer(object):
    def __init__(self, name=None):
        self.name = name

    def __enter__(self):
        self.tstart = time.time()

    def __exit__(self, type, value, traceback):
        if self.name:
            print('[%s]' % self.name,)
        print('Elapsed: %s' % (time.time() - self.tstart))

# Sympify that ignores some built in names
__ignored_sym_func = ['rad', 're']
__ignored_sym = dict((sym, sympy.Symbol(sym)) for sym in __ignored_sym_func)
def sympify(expr, *args, **kwargs):
    """Allows using sympy on expressions with 'reserved' keywords"""
    return sympy.sympify(expr, locals=__ignored_sym, *args, **kwargs)

# Source: http://stackoverflow.com/questions/279561/what-is-the-python-equivalent-of-static-variables-inside-a-function
def static_var(varname, value):
    """Decorator that defines a static variable inside a function"""
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate


def fix_carets(expr):
    """Converts carets to exponent symbol in string"""
    import re as _re
    caret = _re.compile('[\^]')
    return caret.sub('**',expr)


import sympy
__ignored_sym_func = ['rad', 're']
__ignored_sym = dict((sym, sympy.Symbol(sym)) for sym in __ignored_sym_func)


def sympify2(expr, *args, **kwargs):
    """Allows using sympy on expressions with 'reserved' keywords"""
    return sympy.sympify(expr, locals=__ignored_sym, *args, **kwargs)



from .helpers import root
from .beluga import bvp_algorithm, guess_generator
from .continuation import ContinuationList as init_continuation
from .beluga import setup_beluga, solve
from .problem import OCP

import os
import glob

modules = glob.glob(os.path.dirname(__file__)+"/*.py")
__all__ = [ os.path.basename(f)[:-3] for f in modules]



#==============================================================================#
# Author: Justin Mansell (2016)
# Description: Continuation using a heuristic path algorithm (HPA).
#==============================================================================#
from beluga.continuation.HPA_Variable import HPA_Variable
import numpy as np
import itertools
import logging
import copy

#Define a new class for nodes along the continuation path
class node:

    #Constructor function
    def __init__(self, ind, val, par = []):
        self.ind = ind #graph coordinates of the node
        self.par = par #the node's parent node
        self.val = val #values of the continuation variables at the node
        self.g = 0
        self.h = 0
        self.f = 0
        self.bvp = None #BVP object associated with this node

    #Function for printing a node object (for debugging only)
    def __repr__(self):
        return repr((self.ind, self.val, self.f))

class HPA_Strategy(object):
    """Defines one continuation step in continuation set"""
    # A unique short name to select this class
    name = 'HPA'

    def __init__(self, hweight=0.75, max_steps = 100, no_diag = False, vars=[], bvp=None):
        self.bvp = bvp
        self.hweight = hweight #Heuristic weight factor (f = (1-w)*g + w*h)
        self.max_steps = max_steps #Maximum number of continuation steps
        self._num_cases = max_steps
        self.confines = [] #List of confined options
        self.no_diag = no_diag #Allow diagonal moves?
        self.vars = {}  # dictionary of variables
        self.ctr  = 0   # iteration counter
        self.last_bvp = None
        self.last_node = None
        self.convars = [] #Ordered list of coninuation variables
        self.init_vals = [] #List of initial values for continuation variables
        self.target_vals = [] # "    target values  "      "    "       "
        self.graphdims = [] #Number of nodes along the axis of each cont. var.
        self.spacings = [] #Variable spacing (log or linear) for each cont. var.
        self.start_node = None
        self.goal_node = None
        self.goal_attempts = 0 #Number of attempts at solving the goal node
        self.moves = [] #List of possible move directions
        self.open_list = [] #List of frontier nodes
        self.closed_list = [] #List of attempted nodes

    def reset(self):
        """Resets the graph search"""
        self.ctr  = 0
        self.last_bvp = None
        self.last_node = None
        self.convars = []
        self.init_vals = []
        self.target_vals = []
        self.graphdims = []
        self.spacings = []
        self.start_node = None
        self.goal_node = None
        self.goal_attempts = 0
        self.moves = []
        self.open_list = []
        self.closed_list = []

    def clear(self):
        """Clears all the previously set continuation variables"""
        self.vars = {}
        self.reset()

    def set_bvp(self, bvp):
        self.bvp = bvp

        #Graph properites stored in lists to preserve order of the continuation vars.
        convars = [] #List of variable names for the continuation variables
        init_vals = [] #List of initial values for each continuation variable
        graphdims = [] #List of graph dimenions
        spacings = [] #List of parameter spacing for each dimension
        target_vals = [] #List of target values for each continuation variable
        #confined = [] #Which variables are confined to the grid

        # Initialize strategy parameters
        for var_type in self.vars.keys():
            for var_name in self.vars[var_type].keys():
                # Look for the variable name from continuation in the BVP
                if var_name not in bvp.solution.aux[var_type].keys():
                    raise ValueError('Variable '+var_name+' not found in boundary value problem')

                #Load continuation variable data
                self.convars.append(var_name)
                self.init_vals.append(bvp.solution.aux[var_type][var_name])
                self.graphdims.append(self.vars[var_type][var_name].nodes)
                self.spacings.append(self.vars[var_type][var_name].spacing)
                self.target_vals.append(self.vars[var_type][var_name].target)
                self.confines.append(self.vars[var_type][var_name].confined)

        self.start_node = node(np.zeros(len(self.graphdims)),self.init_vals)
        self.start_node.bvp = bvp #Initialize the bvp
        self.goal_node = node(self.graphdims-np.ones(len(self.graphdims)),self.target_vals) #Might be unnecessary
        self.open_list = []
        self.last_node = self.start_node

        #Warning
        if len(self.graphdims)<2: raise ValueError('HPA continuation not appropriate for continuation on a single variable. Use manual or bisection strategy instead.')

        #Generate the possible move direcitons
        moves = list(itertools.product([-1,0,1],repeat=len(self.graphdims))) #Generate possible directions
        moves.remove(tuple([0]*len(self.graphdims))) #Get rid of the 0 vector as a possible direction
        if self.no_diag is True: moves = [d for d in moves if not np.linalg.norm(d)>1] #Get rid of diagonal directions
        self.moves = moves

    def set(self, var_type,name,target,nodes,spacing='linear',confined=False):
        if var_type not in self.vars.keys():
            self.vars[var_type] = {}

        # Create continuation variable object
        self.vars[var_type][name] = HPA_Variable(name,target,nodes,spacing,confined)
        return self

    def terminal(self, name,target,nodes,spacing='linear',confined=False):
        self.set('terminal',name,target,nodes,spacing,confined)
        return self

    def initial(self, name,target,nodes,spacing='linear',confined=False):
        self.set('initial',name,target,nodes,spacing,confined)
        return self

    def const(self, name,target,nodes,spacing='linear',confined=False):
        self.set('const',name,target,nodes,spacing,confined)
        return self

    # def constraint(self, name,target):
    #     self.set('constraint',name,target)
    #     return self

    def __iter__(self):
        """Define class as being iterable"""
        return self

    def __next__(self):
        return self.next()

    #Basically a dummy function to make Beluga.py happy. We don't know what
    #the steps will be a-priori in graph search.
    def num_cases(self,num_cases=None, spacing='linear'):
        if num_cases is None:
            return self._num_cases
        else:
            if self.ctr > 0:
                raise RuntimeError('Cannot set num_cases during iteration')

            self._num_cases = num_cases
            self._spacing   = spacing
            return self

    #Cost to move between nodes
    def PathCost(self, ni, nf):
        """The incremental cost to go from one node to an adjacent node"""
        #ni, nf = node objects
        return 1.0 #Path cost is just the cardinality length between nodes

    #Heuristic function
    def Heuristic(self, n):
        """Expected cost to reach the goal node from node n"""
        #Cardinal length to reach goal, plus a small constribution of the actual
        #distance to favor more obvious choices in the case of a tie between nodes.
        return max(abs(n.ind - self.goal_node.ind))+ \
                                0.01*np.linalg.norm(n.ind-self.goal_node.ind)

    def ind2vals(self, ind):
        """"Converts a node index into values for the continuation variables"""
        vals = []
        for i,var in enumerate(self.convars):
            if self.spacings[i] == 'linear':
                vals.append((self.target_vals[i] - self.init_vals[i])* \
                                  ind[i]/(self.graphdims[i]-1) + self.init_vals[i])

            if self.spacings[i] == 'log':
                vals.append(self.init_vals[i]*(self.target_vals[i]/ \
                                self.init_vals[i])**(ind[i]/(self.graphdims[i]-1)))

        return vals

    def next(self, ignore_last_step = False):
        """Generator class to create BVPs for the continuation step iterations
        """

        if self.bvp is None:
            raise ValueError('No boundary value problem associated with this object')

        if max(abs(self.last_node.ind - self.goal_node.ind)) == 0 and self.last_node.bvp.solution.converged:
            raise StopIteration #Solved original problem!

        if self.ctr >= self.max_steps:
            raise ValueError('Exceeded maximum number of continuation steps')

        #Check convergence of last bvp
        if not ignore_last_step and self.last_bvp is not None \
            and not self.last_bvp.solution.converged and \
            max(abs(self.last_node.ind - self.goal_node.ind)) != 0:

            #NOTE:If the node we attempted was the goal node, we don't want to
            #add it to closed_list, otherwise HPA would never attempt it again.
            #We already popped the goal node off of open_list, so if we do nothing
            #HPA will explore a subproblem adjacent to the goal node. In doing so,
            #HPA will generate the goal node again as a successor, but this time
            #the goal node will have a different parent (and thus have a new chance)
            #at being solved.

            #Problem diverged, so we don't want to expand any successors from it
            self.closed_list.append(self.last_node)

        #Last problem converged or goal node was attempted but failed; generate new nodes
        else:
            if max(abs(self.last_node.ind - self.goal_node.ind)) != 0:
                self.closed_list.append(self.last_node) #Don't close the goal
            else: self.goal_attempts += 1 #Goal attempted but failed

            #Generate successors to last_node
            successors = []
            for direction in self.moves:
                new_ind = self.last_node.ind + direction
                new_vals = self.ind2vals(new_ind)
                successors.append(node(new_ind,new_vals,self.last_node))

            for successor in successors:
                skip = 0 #skip flag

                #Compute the value function of each successor
                successor.g = self.last_node.g + self.PathCost(self.last_node,successor)
                successor.h = self.Heuristic(successor)
                successor.f = (1-self.hweight)*successor.g + self.hweight*successor.h

                #Make sure node is not in open_list or closed_list already
                for n in self.open_list:
                    if max(abs(n.ind - successor.ind)) == 0:
                        skip = 1
                        break
                for n in self.closed_list:
                    if max(abs(n.ind - successor.ind)) == 0:
                        skip = 1
                        break

                #Make sure node is not outside of the grid, if this isn't allowed
                for ii,grid_confine in enumerate(self.confines):
                    if grid_confine == True:
                        if successor.ind[ii]>self.graphdims[ii]-1 or successor.ind[ii]<0: skip=1

                #Add successors to the frontier and assign bvp objects
                if skip == 1:
                    continue #Check to see if we need to skip this successor
                else:
                    #Assign a bvp object to each valid successor
                    successor.bvp = copy.deepcopy(successor.par.bvp)
                    for var_type in self.vars:
                        for var_name in self.vars[var_type]:
                            successor.bvp.solution.aux[var_type][var_name] = successor.val[self.convars.index(var_name)]
                    self.open_list.append(successor) #Add successor to the frontier
        #We have now updated the frontier with successors to the last bvp

        #Generate a new bvp from a node on the frontier
        if not self.open_list:
            raise ValueError('No new nodes left to explore')
        if self.goal_attempts >= len(self.moves):
            raise ValueError('Surrounded the goal node but could not solve it')

        self.open_list = sorted(self.open_list, key=lambda n: n.f, reverse=True)
        q = self.open_list.pop() #Pick the node with the smallest f
        print('Attemping node:',q.ind,'from node',q.par.ind)
        print('Node values:',self.convars,q.val)
        print('Attempts at the goal:',self.goal_attempts)

        #Clean up and return the new bvp
        self.ctr += 1
        self.last_node = q
        self.bvp = q.bvp #Update the bvp
        self.last_bvp = self.bvp
        return self.bvp

        #IDEA:
        #Show the costate guesses in between steps to see how much the solution has changed.



import numpy as np
import logging
import inspect
import functools
import sys


class ActivateConstraint(object):
    """
    Continuation class activating a constraint.
    """
    strategy_name = 'activate_constraint'

    def __init__(self, name):
        self.name = name
        self.solved = False
        self.ctr  = 0

    def num_cases(self):
        return 1

    def reset(self):
        self.solved = False

    def clear(self):
        self.solved = False

    def init(self, sol, problem_data):
        """Split the solution into arcs to introduce constraint."""
        # Get expr corresponding to name
        s = sol.aux['constraints'][self.name]
        expr = s['expr']
        arc_type = s['arc_type']
        pi_list = s['pi_list']
        num_params = problem_data['num_params']

        current_arcs = sol.arcs
        current_arcseq = sol.aux.get('arc_seq', (0,))
        if current_arcs is None:
            current_arcs = [(0, len(sol.t)-1)]

        sol.prepare(problem_data, 200, True)
        # Evaluate expr on sol
        s_vals = sol.evaluate(expr)

        # Find zero crossing
        # TODO: add check to see if constraint active

        if s['direction'] == '>':
            # Find min value (most away from zero)
            s_lim_i = np.argmin(s_vals)
            s_lim_val = s_vals[s_lim_i]
            if s_lim_val > 0:
                print('Constraint is not active')
            s_lim_val = min(0, s_lim_val)
            # print('Max_violation : ', s_lim_val)
        else:
            # Find max value
            s_lim_i = np.argmax(s_vals)
            s_lim_val = s_vals[s_lim_i]
            if s_lim_val < 0:
                print('Constraint is not active')
            s_lim_val = max(0, s_lim_val)
            # print('Max_violation : ', s_lim_val)

        if s_lim_i == 0:
            s_lim_i = 1
        elif s_lim_i == len(sol.x)-1:
            s_lim_i = s_lim_i - 1

        arc_to_split = int(np.floor(sol.x[s_lim_i]))
        logging.info('Added constrained arc with max violation : %.4lf %s in arc %d' % (s_lim_val, s['unit'], arc_to_split))

        # Store constraint limit in aux
        sol.aux['constraint'][(self.name, 1)] = s_lim_val

        arc_num = arc_to_split
        old_arc_idx = sol.arcs[arc_num] # See where previous arc starts and ends

        # new arc index relative to start of previous arc
        idx_arc_start = s_lim_i - old_arc_idx[0]
        idx_arc_end = s_lim_i+1 - old_arc_idx[0]

        old_arc_idx = (old_arc_idx[0], old_arc_idx[1]+1)
        old_arc_t = np.array(sol.x[slice(*old_arc_idx)], copy=True)
        old_arc_y = np.array(sol.y[:,slice(*old_arc_idx)], copy=True)
        old_arc_u = np.array(sol.u[:,slice(*old_arc_idx)], copy=True)
        original_tf = old_arc_y[-1,0] # TODO: We don't put time in as a state anymore.

        t_before = old_arc_t[idx_arc_start]*original_tf
        t_during = (old_arc_t[idx_arc_end] - old_arc_t[idx_arc_start])*original_tf
        t_after = (old_arc_t[-1] - old_arc_t[idx_arc_end])*original_tf

        logging.info('Arc position : t='+str(sol.x[s_lim_i])+'s')
        old_arc_t[0:idx_arc_start+1] = (old_arc_t[0:idx_arc_start+1] - old_arc_t[0])/(old_arc_t[idx_arc_start] - old_arc_t[0]) + arc_num # TODO: Fix for multi arc
        old_arc_t[idx_arc_end:] = (old_arc_t[idx_arc_end:] - old_arc_t[idx_arc_end])/(old_arc_t[-1] - old_arc_t[idx_arc_end]) + arc_num + 2

        new_arc_t = np.hstack((old_arc_t[:idx_arc_start+1], old_arc_t[idx_arc_start:idx_arc_end+1], old_arc_t[idx_arc_end:]))
        new_arc_y = np.hstack((old_arc_y[:,:idx_arc_start+1], old_arc_y[:,idx_arc_start:idx_arc_end+1], old_arc_y[:,idx_arc_end:]))
        new_arc_u = np.hstack((old_arc_u[:,:idx_arc_start+1], old_arc_u[:,idx_arc_start:idx_arc_end+1], old_arc_u[:,idx_arc_end:]))

        new_arc_y[-1,:idx_arc_start+1] = t_before
        new_arc_y[-1,idx_arc_start+1] = t_during # "tf" for constrained arc
        new_arc_y[-1,idx_arc_end+1] = t_during
        new_arc_y[-1,idx_arc_end+2:] = t_after

        num_odes = sol.y.shape[0]
        num_controls = sol.u.shape[0]
        sol.t = np.hstack((sol.t[:old_arc_idx[0]], new_arc_t, sol.t[old_arc_idx[1]:]))
        sol.y = np.hstack((sol.y[:,:old_arc_idx[0]], new_arc_y, sol.y[:,old_arc_idx[1]:]))
        sol.u = np.hstack((sol.u[:,:old_arc_idx[0]], new_arc_u, sol.u[:,old_arc_idx[1]:]))

        new_arcs = [(old_arc_idx[0], old_arc_idx[0]+idx_arc_start),
                    (old_arc_idx[0]+idx_arc_start+1, old_arc_idx[0]+idx_arc_end+1),
                    (old_arc_idx[0]+idx_arc_end+2, old_arc_idx[0]+len(new_arc_t)-1)]

        sol.arcs = (*sol.arcs[:arc_num], *new_arcs, *sol.arcs[arc_num+1:])
        arc_seq = (*sol.aux['arc_seq'][:arc_num+1], arc_type, *sol.aux['arc_seq'][arc_num:])
        sol.aux['arc_seq'] = arc_seq

        pi_idx_start = len(sol.parameters)
        pi_idx = np.array(list(range(pi_idx_start, pi_idx_start+len(pi_list))))
        sol.parameters = np.append(sol.parameters, np.ones(len(pi_list))*0.0)

        if len(pi_idx) == 0:
            pi_idx = None
        new_pi_seq = (None, pi_idx, None)
        pi_seq = (*sol.aux['pi_seq'][:arc_num], *new_pi_seq, *sol.aux['pi_seq'][arc_num+1:])
        sol.aux['pi_seq'] = pi_seq

        # keyboard()
        self.sol = sol

    def __iter__(self):
        """Define class as being iterable"""
        return self

    def __next__(self):
        if self.solved:
            raise StopIteration
        else:
            self.solved = True
            self.ctr += 1
            return self.sol.aux


class ContinuationList(list):
    def __init__(self):
        # Create list of available strategies
        clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
        self.strategy_list = {obj.strategy_name: obj
                                for (name,obj) in clsmembers
                                if hasattr(obj,'strategy_name')
                              }

    def add_step(self, strategy='manual', *args, **kwargs):
        """
        Adds a continuation step with specified strategy

        Returns the strategy object to further chain calls into it
        """
        # Create object if strategy is specified as a string
        if isinstance(strategy, str):
            if strategy in self.strategy_list:
                strategy_obj = self.strategy_list[strategy](*args, **kwargs)
            else:
                logging.error('Invalid strategy name')
                raise ValueError('Continuation strategy : '+strategy+' not found.')
        else:
            strategy_obj = strategy

        self.append(strategy_obj)
        return strategy_obj


class ContinuationVariable(object):
    """
    Class containing information for a continuation variable.
    """

    def __init__(self, name, target):
        self.name = name
        self.target = target
        self.value = np.nan
        self.steps = []


# Can be subclassed to allow automated stepping
class ManualStrategy(object):
    """
    Class defining the manual continuation strategy.
    """
    # A unique short name to select this class
    strategy_name = 'manual'

    def __init__(self, num_cases = 1,vars=[], sol=None):
        self.sol = sol
        self._num_cases = num_cases
        self._spacing = 'linear'
        self.vars = {}  # dictionary of values
        self.ctr  = 0   # iteration counter
        self.last_sol = None

        self.terminal = functools.partial(self.set, param_type='terminal')
        self.initial = functools.partial(self.set, param_type='initial')
        self.const = functools.partial(self.set, param_type='const')
        self.orig_num_cases = num_cases
        self.constant = self.const

    def reset(self):
        """Resets the internal step counter to zero"""
        self.ctr = 0
        self.last_sol = None

    def clear(self):
        """Clears all the previously set continuation variables"""
        self.vars = {}
        self.reset()

    def constraint(self, name, target, index):
        """Continuation on constraint limit.

        index : Specify which constraint arc to use."""
        # sol.aux['constraints'][self.name]['limit'][1]
        self.set((name,index), target, param_type='constraint')

    def var_iterator(self):
        for var_type in self.vars.keys():
            for var_name in self.vars[var_type].keys():
                yield var_type, var_name

    # TODO: Change to store only stepsize and use yield
    def init(self, sol, problem_data):
        self.sol = sol

        # Iterate through all types of variables
        for var_type, var_name in self.var_iterator():
        # for var_type in self.vars.keys():
        #     for var_name in self.vars[var_type].keys():
            # Look for the variable name from continuation in the BVP
            if var_name not in sol.aux[var_type].keys():
                raise ValueError('Variable '+var_name+' not found in boundary value problem')

            # Set current value of each continuation variable
            self.vars[var_type][var_name].value = sol.aux[var_type][var_name]
            # Calculate update steps for continuation process
            if self._spacing == 'linear':
                self.vars[var_type][var_name].steps = np.linspace(self.vars[var_type][var_name].value,
                                                                  self.vars[var_type][var_name].target,
                                                                  self._num_cases)
            elif self._spacing == 'log':
                self.vars[var_type][var_name].steps = np.logspace(np.log10(self.vars[var_type][var_name].value),
                                                                  np.log10(self.vars[var_type][var_name].target),
                                                                  self._num_cases)

    def set(self,name,target,param_type):
        """
        Sets the target value for the specified parameter
        """
        if param_type not in self.vars.keys():
            self.vars[param_type] = {}

        # Create continuation variable object
        self.vars[param_type][name] = ContinuationVariable(name,target)
        return self

    def num_cases(self,num_cases=None,spacing='linear'):
        if num_cases is None:
            return self._num_cases
        else:
            if self.ctr > 0:
                raise RuntimeError('Cannot set num_cases during iteration')

            self._num_cases = num_cases
            self.orig_num_cases = num_cases
            self._spacing   = spacing
            return self

        self.set('const', name, target)
        return self

    def __str__(self):
        return str(self.vars)

    def __iter__(self):
        """Define class as being iterable"""
        return self

    def __next__(self):
        return self.next()

    def next(self, ignore_last_step = False):
        """Generator class to create BVPs for the continuation step iterations
        ignore_last_step: Should the non-convergence of previous step be ignored?
        """

        if self.sol is None:
            raise ValueError('No boundary value problem associated with this object')

        if not ignore_last_step and self.last_sol is not None and not self.last_sol.converged:
            logging.error('The last step did not converge!')
            raise RuntimeError('Solution diverged! Stopping.')

        if self.ctr >= self._num_cases:
            raise StopIteration

        # Update auxiliary variables using previously calculated step sizes
        total_change = 0.0
        for var_type in self.vars:
            for var_name in self.vars[var_type]:
                self.sol.aux[var_type][var_name] = self.vars[var_type][var_name].steps[self.ctr]
                total_change += abs(self.vars[var_type][var_name].steps[self.ctr])

        self.ctr += 1
        self.last_sol = self.sol
        return self.sol.aux


class BisectionStrategy(ManualStrategy):
    """
    Defines the bisection continuation strategy.
    """
    strategy_name = 'bisection'

    def __init__(self, initial_num_cases = 5, max_divisions=10, num_divisions = 2, vars=[], aux=None):
        super(BisectionStrategy, self).__init__(num_cases=initial_num_cases)
        self.last_sol = None
        self.num_divisions = num_divisions
        self.max_divisions = max_divisions
        self.division_ctr = 0
        self.orig_num_cases = initial_num_cases

    def __str__(self):
        return str(self.vars)

    def next(self):
        """Generator class to create BVPs for the continuation step iterations

            last_converged: Specfies if the previous continuation step converged
        """
        # If it is the first step or if previous step converged
        # continue with usual behavior
        if self.ctr == 0 or self.last_sol.converged:
            # Reset division counter
            self.division_ctr = 0
            return super(BisectionStrategy, self).next()

        # If first step didn't converge, the process fails
        if self.ctr == 1:
            logging.error('Initial guess should converge for automated continuation to work!!')
            raise RuntimeError('Initial guess does not converge.')
            return self.last_sol

        if self.division_ctr > self.max_divisions:
            logging.error('Solution does not without exceeding max_divisions : '+str(self.max_divisions))
            raise RuntimeError('Exceeded max_divisions')

        # If previous step did not converge, move back a half step
        for var_type, var_name in self.var_iterator():
            # Set current value of each continuation variable
            # self.vars[var_type][var_name].value = aux[var_type][var_name]
            # insert new steps
            old_steps = self.vars[var_type][var_name].steps
            if self._spacing == 'linear':
                new_steps = np.linspace(old_steps[self.ctr-2],old_steps[self.ctr-1],self.num_divisions+1)
            elif self._spacing == 'log':
                new_steps = np.logspace(np.log10(old_steps[self.ctr-2]),np.log10(old_steps[self.ctr-1]),self.num_divisions+1)
            else:
                raise ValueError('Invalid spacing type')

            # Insert new steps
            self.vars[var_type][var_name].steps = np.insert(
                    self.vars[var_type][var_name].steps,
                    self.ctr-1,
                    new_steps[1:-1] # Ignore first element as it is repeated
                )
        # Move the counter back
        self.ctr = self.ctr - 1
        # Increment total number of steps
        self._num_cases = self._num_cases + self.num_divisions-1

        logging.info('Increasing number of cases to '+str(self._num_cases)+'\n')
        self.division_ctr = self.division_ctr + 1
        return super(BisectionStrategy, self).next(True)


class HPA_Variable(object):
    def __init__(self,name,target,Nnodes,spacing='linear',confined=False):
        self.name = name
        self.target = target
        self.spacing = spacing #spacing of the continuation parameter
        self.confined = confined #whether or not the search can venture beyond the limits of the grid
        self.nodes = Nnodes #number of nodes along this axis of the graph
        self.value = np.nan
        self.index = np.nan # Index of state in BVP, obsolete
        self.steps = []



"""Quasi-linear Chebyshev Picard Iteration method"""

from beluga.ivpsol.integrators.mcpi import mcpi, mcpi_init, absdiff
from numba import njit, prange, jit
from .BaseAlgorithm import BaseAlgorithm
import pystache
import simplepipe as sp
import numpy as np
import functools as ft
import imp
from beluga.problem import BVP
import sys
import logging
# def create_odefn(problem_data, module):
#

def create_module(problem_data):
    """Creates a new module for storing compiled code.

    Parameters
    ----------
    problem_data - dict
        Problem data dictionary

    Returns
    -------
    New module for holding compiled code
    """
    problem_name = problem_data['problem_name']
    module = imp.new_module('_beluga_'+problem_name)
    return module

def load_eqn_template(problem_data, template_file,
                        renderer = pystache.Renderer(escape=lambda u: u)):
    """Loads pystache template and uses it to generate code.

    Parameters
    ----------
        problem_data - dict
            Workspace defining variables for template

        template_file - str
            Path to template file to be used

        renderer
            Renderer used to convert template file to code

    Returns
    -------
    Code generated from template
    """
    template_path = beluga.root()+'/optimlib/templates/'+problem_data['method']+'/qcpi/'+template_file
    with open(template_path) as f:
        tmpl = f.read()
        # Render the template using the data
        code = renderer.render(tmpl, problem_data)
        return code

def compile_code_py(code_string, module, function_name):
    """
    Compiles a function specified by template in filename and stores it in
    self.compiled

    Parameters
    ----------
    code_string - str
        String containing the python code to be compiled

    module - dict
        Module in which the new functions will be defined

    function_name - str
        Name of the function being compiled (this must be defined in the
        template with the same name)

    Returns:
        Compiled function
    """
    # For security
    module.__dict__.update({'__builtin__':{}})
    exec(code_string, module.__dict__)
    return getattr(module,function_name)

QCPICodeGen = sp.Workflow([
    # Create module for holding compiled code
    sp.Task(ft.partial(create_module), inputs='problem_data', outputs=('code_module')),

    # sp.Task(make_functions, inputs=('problem_data', 'code_module'), outputs=('code_module','compute_control_fn')),
    # Load equation template files and generate code
    sp.Task(ft.partial(load_eqn_template,
                template_file='deriv_func.py.mu'),
            inputs='problem_data',
            outputs='deriv_func_code'),
    sp.Task(ft.partial(load_eqn_template,
                template_file='bc_func.py.mu'),
            inputs='problem_data',
            outputs='bc_func_code'),

    # Compile generated code
    sp.Task(ft.partial(compile_code_py, function_name='deriv_func'),
            inputs=['deriv_func_code', 'code_module'],
            outputs='deriv_func_fn'),
    sp.Task(ft.partial(compile_code_py, function_name='bc_func'),
            inputs=['bc_func_code', 'code_module'],
            outputs='bc_func_fn'),
], description='Generates and compiles the required BVP functions from problem data')


pert_eom_bck = None
@ft.lru_cache(maxsize=16)
def make_pert_eom(nOdes, q, eom_fn, executor):
    """Makes EOM that evaluates perturbed & unperturbed states at all time steps"""

    @njit
    def pert_eom_inner(idx, all_args):
        t, X, dXdt, args = all_args
        i, j = idx
        X_ = X[:,i*nOdes:(i+1)*nOdes]
        dXdt_ = dXdt[:,i*nOdes:(i+1)*nOdes]
        dXdt_[j,:] = eom_fn(t[j], X_[j,:nOdes], X_[j,nOdes:nOdes], *args)

    def pert_eom(t, X, dXdt, *args):
        # executor.map(ft.partial(pert_eom_inner, all_args=(t, X, dXdt, args)),
        #              np.ndindex(q+2, len(t)))
        # for i,j in np.ndindex(q+2, len(t)):
        #     pert_eom_inner((i, j), (t, X, dXdt, args))

        # for i in range(q+2):
        #     # Extract only relevant states
        #     X_ = X[:,i*nOdes:(i+1)*nOdes]
        #     dXdt_ = dXdt[:,i*nOdes:(i+1)*nOdes]
        #     # One timestep at a time
        #     for j in range(len(t)):
        #         dXdt_[j,:] = eom_fn(t[j], X_[j,:nOdes], X_[j,nOdes:nOdes], *args)
        #         # dXdt_[{{num_states}}+{{dae_var_num}}:]
        #         # print(X_[j,:].shape, dXdt_[j,:].shape, retval.shape)
        #         # eom_fn(t[j], X_[j,:], dXdt_[j,:], *args)

    global pert_eom_bck
    pert_eom_bck = pert_eom
    return pert_eom

@ft.lru_cache(maxsize=16)
def make_mcpi_eom(eom_fn):
    """Makes EOM that evaluates perturbed & unperturbed states at all time steps"""

    @njit(parallel=False)
    def mcpi_eom(t, X_, dXdt_, *args):
        # One timestep at a time
        for j in range(len(t)):
            eom_fn(t[j], X_[j,:], dXdt_[j,:], *args)

    return mcpi_eom


@ft.lru_cache(maxsize=16)
def make_bc_jac(bcfn, nOdes, step_size=1e-6):
    """Makes compiled BC jacobian function using forward difference"""

    @jit(parallel=True)
    def jac_fn(Yb, jac, *args):
        fx = bcfn(Yb, *args)
        steps = np.eye(nOdes)*step_size + Yb[:nOdes]
        for i in prange(nOdes):
            fxh = bcfn(steps[i], *args)
            jac[:,i] = (fxh - fx)/step_size

        return fx

    return jac_fn


@ft.lru_cache(maxsize=16)
def make_perf_idx(bc_left_fn, bc_right_fn):
    @jit(parallel=True)
    def perf_idx(x_t, A_t, alpha, *args):
        P = 0
        # MCPI gives reversed time history
        x_f = x_t[0,:] + alpha*A_t[0,:]
        x_0 = x_t[-1,:] + alpha*A_t[-1,:]

        f_vec = bc_left_fn(x_0, *args)
        g_vec = bc_right_fn(x_f, *args)
        P = P + f_vec.T @ f_vec + g_vec.T @ g_vec
        return P
    return perf_idx


import concurrent.futures

class QCPI_Parallel(BaseAlgorithm):
    def __init__(self, tolerance=1e-6, max_iterations=100, max_error=10, N=21, verbose=True):
        self.tolerance = tolerance
        self.max_iterations = max_iterations
        self.verbose = verbose
        self.max_error = max_error
        self.N = N
        self.saved_deriv_func = False
        self.executor = concurrent.futures.ProcessPoolExecutor()

    def close(self):
        self.executor.shutdown()

    def preprocess(self, problem_data):
        """Code generation and compilation before running solver."""

        out_ws = QCPICodeGen({'problem_data': problem_data})
        print(out_ws['bc_func_code'])
        print(out_ws['deriv_func_code'])

        self.bvp = BVP(out_ws['deriv_func_fn'],
                       out_ws['bc_func_fn'], None) #out_ws['compute_control_fn'])#out_ws['compute_control_fn'])

        self.stm_ode_func = self.bvp.deriv_func
        self.module = out_ws['code_module']
        self.left_bc_mask = out_ws['code_module'].bc_free_mask
        self.num_dae_vars = out_ws['code_module'].num_dae_vars
        # self.left_bc_mask = problem_data['bc_free_mask']
        # print(self.left_bc_mask)
        self.bc_left_fn = out_ws['code_module'].bc_func_left
        self.bc_right_fn  = out_ws['code_module'].bc_func_right
        # self.deriv_func = njit(parallel=True)(out_ws['code_module'].deriv_func_mcpi)
        # if deriv_func_bck is not None:
        #     print('Loaded deriv_func from cache')
        #     self.deriv_func = deriv_func_bck
        #     self.saved_deriv_func = True
        # else:
        # self.deriv_func = out_ws['code_module'].deriv_func_mcpi
        self.deriv_func = out_ws['code_module'].deriv_func

        self.mcpi_eom = make_mcpi_eom(self.deriv_func)
        sys.modules['_beluga_'+problem_data['problem_name']] = out_ws['code_module']
        return out_ws['code_module']

    def solve(self,solinit):
        x  = solinit.x
        # Get initial states from the guess structure
        y0g = solinit.y[:,0]
        paramGuess = solinit.parameters

        deriv_func = self.bvp.deriv_func
        bc_func = self.bvp.bc_func

        aux = solinit.aux
        # Only the start and end data are required for ode45
        arcs = solinit.arcs

        ya = solinit.y[:,0]
        yb = solinit.y[:,-1]

        # # Set initial values
        icvals= np.array(list(aux['initial'].values()))
        icval_slice = np.array(self.left_bc_mask[:-self.num_dae_vars])==0
        ya[:-self.num_dae_vars][icval_slice] = icvals[icval_slice]

        # Extract number of ODEs in the system to be solved
        nXandLam = y0g.shape[0]
        nParams = paramGuess.size
        nOdes = nXandLam + nParams

        ## Setup QCPI
        x_0 = np.concatenate((ya, paramGuess))

        res_left = self.bc_left_fn(x_0, aux)
        left_jac = np.zeros((len(res_left), nOdes))
        left_bc_jac_fn = make_bc_jac(self.bc_left_fn, nOdes)
        left_bc_jac_fn(x_0, left_jac, aux)


        # y_pert = np.absolute(np.max(solinit.y, axis=1))*0.000001
        # y_pert = np.append(y_pert, np.absolute(solinit.parameters)*0.000001)
        # y_pert[y_pert < 1*min(self.tolerance,1e-6)] = 1*min(self.tolerance,1e-6)
        # A_j0 = np.vstack((np.zeros(nOdes), y_pert*np.eye(nOdes, dtype=np.float64)))
        # Each row is one initial condition for particular solution
        # A_j0 = np.eye(nOdes)
        # A_j0[np.diag_indices(nOdes)] = self.left_bc_mask
        # A_j0 = np.unique(A_j0, axis=0)*1e-6   # Remove duplicates
        A_j0 = np.vstack((np.zeros(nOdes), np.eye(nOdes, dtype=np.float64)*1e-6))

        # A_j0 = np.vstack((np.zeros(nOdes), 0.01*np.eye(nOdes)))

        q = len(A_j0) - 1
        # from beluga.utils import keyboard
        # keyboard()
        # Set up perturbed ICs
        xp_0 = np.hstack((x_0, A_j0.flat)) # Add perturbed ICs as extra states
        xp_0[nOdes:] = np.tile(xp_0[:nOdes], (q+1,)) + A_j0.flat  # Add perturbations to ICs
        xpm_0 = np.reshape(xp_0, (q+2, nOdes))  # Matrix 'view' of xp_0
        # tspan = x
        tspan = x[0], x[-1]

        t_short = [tspan[0], tspan[-1]]
        converged = False
        N = self.N
        max_iter = self.max_iterations

        # Setup MCPI
        w1 = (tspan[-1]-tspan[0])/2
        w2 = (tspan[-1]+tspan[0])/2
        tau, C_x, C_a = mcpi_init(N)      # tau is in reverse order (1 to -1)
        C_a = w1 * C_a
        t_arr = tau*w1 + w2
        # Make functions
        pert_eom = make_pert_eom(nOdes, q, self.deriv_func, self.executor)
        bc_jac_fn = make_bc_jac(self.bc_right_fn, nOdes)
        perf_idx_fn = make_perf_idx(self.bc_left_fn, self.bc_right_fn)

        const = tuple(aux['const'].values())

        try:
            _, x_guess2 = mcpi(pert_eom, tspan, xp_0, N=N, args=(const,))
        except:
            x_guess2 = None

        # x_guess2 = None
        # if solinit.extra is None or solinit.extra.shape[0] != (N+1):
        #     try:
        #         if solinit.extra is not None:
        #             print('Propagating eqns as',solinit.extra.shape[0],'!=',N+1)
        #         else:
        #             print('solinit.extra is None')
        #         _, x_guess2 = mcpi(pert_eom, tspan, xp_0, N=N, args=(const,))
        #     except:
        #         x_guess2 = None
        if x_guess2 is None or np.isnan(x_guess2.flat).any():
            if solinit.extra is not None and not np.isnan(solinit.extra.flat).any():
                x_guess = solinit.extra
            else:
                x_guess = np.tile(xp_0, (N+1, 1))  # Each column -> time history of one state
        else:
            x_guess = np.flipud(x_guess2)
            # print('nsoododoshds')

        x_guess_orig = np.copy(x_guess)
        x0_twice = 2*x_guess[-1]
        g_arr = np.empty_like(x_guess)

        res = self.bc_right_fn(x_0, aux)

        nBCs = len(res)
        psi_jac = np.zeros((nBCs, nOdes))

        alpha1 = 1
        alpha2 = 1
        r0 = None

        err1 = 1000
        err0 = 9999
        np.set_printoptions(precision=4, linewidth=160)
        ctr = 0
        total_iter = 0
        while ctr < max_iter:
            try:
                pert_eom(t_arr, x_guess, g_arr, const)
            except Exception as e:
                logging.error('Perturbed EOM failed.:',e)
                break

            # if np.any(np.isnan(g_arr)):
            #     print('NaaaaaNaaaa')
            #     from beluga.utils import keyboard
            #     keyboard()
            #     break

            beta = (C_a @ g_arr)
            beta[0]+= x0_twice          # Compute Chebyshev coefficients of solution
            x_new = C_x @ beta          # Compute solution

            err1 = np.max(absdiff(x_new, x_guess))
            # print(err1)
            # if err1 > self.max_error:
            #     x_new = x_guess
            #     logging.error('Error exceeded max error')
            #     break
            if np.isnan(err1):
                x_new = x_guess_orig
                print('NaaaaaN')
                break
                from beluga.utils import keyboard
                keyboard()

            if total_iter > 5000:
                print('Total integration iterations exceeded!')
                x_new = x_guess_orig
                break
            if err1 < 1*min(self.tolerance,1e-4) or abs(err0-err1)<min(self.tolerance,1e-4):
                x_tf = x_new[0,:]       # x_new is reverse time history
                res = bc_jac_fn(x_tf[:nOdes], psi_jac, aux) # Compute residue and jacobian
                res_norm_0 = np.amax(np.abs((res)))

                x_t0 = x_new[-1,:]
                res_left = left_bc_jac_fn(x_t0, left_jac, aux)

                # if ctr > -90:
                print('Residual : '+str(res_norm_0))

                if res_norm_0 > self.max_error:
                    x_new = x_guess_orig # Reset to old version in case of NaN
                    logging.error('Error exceeded max error')
                    break
                if ctr > 0 and res_norm_0 < self.tolerance and np.amax(np.abs(res_left)) < self.tolerance:
                    converged = True
                    print('Converged in %d iterations.' % ctr)
                    break

                if np.isnan(res).any():
                    x_new = x_guess_orig # Reset to old version in case of NaN
                    # from beluga.utils import keyboard
                    # keyboard()
                    print('NaaaaaN')
                    break
                # Convert perturbed terminal conditions to matrix
                # Num of rows = q+1, num of cols = num_states
                A_jf = np.reshape(x_tf[nOdes:], (q+1, nOdes))
                # Subtract unperturbed terminal state to get perturbations
                A_jf = (A_jf - x_tf[:nOdes]).T

                # lhs = np.vstack((np.ones((1,q+1)), psi_jac @ A_jf))
                # try:
                #     k_j = np.linalg.solve(lhs, np.hstack((1, -res)))
                # except:
                #     k_j, *_ = np.linalg.lstsq(lhs, np.hstack((1, -res)))


                A0 = np.reshape(x_t0[nOdes:], (q+1, nOdes))
                A0 = (A0 - x_t0[:nOdes]).T

                lhs = np.vstack((np.ones((1,q+1)), psi_jac @ A_jf, left_jac @ A0))

                # First row is all ones, rows after = Jac @ perturbations at tf
                try:
                    k_j = np.linalg.solve(lhs, np.hstack((1, -res, -res_left)))
                except:
                    k_j, *_ = np.linalg.lstsq(lhs, np.hstack((1, -res, -res_left)))
                A_0 = k_j @ A_j0

                # stepsize calculation
                x_t = x_new[:,:nOdes]
                A_t = np.zeros_like(x_t)
                P0 = perf_idx_fn(x_t, A_t, 0.0, aux)
                # if P0 < self.tolerance and ctr > 0:
                #     converged = True
                #     print('Converged (w/ P0) in %d iterations.' % ctr)
                #     break

                # Compute A for every time step
                for i in range(x_guess.shape[0]):
                    x_ti = x_new[i,:]
                    A_ji = np.reshape(x_ti[nOdes:], (q+1, nOdes))
                    # Subtract unperturbed x(t) to get perturbation at time t
                    A_ji = (A_ji - x_ti[:nOdes])
                    A_ti = k_j @ A_ji
                    A_t[i,:] = k_j @ A_ji

                alpha = 2.0
                P = P0 + 1
                while P > P0 and alpha > min(self.tolerance,1e-4):
                    alpha = alpha/2.0
                    P = perf_idx_fn(x_t, A_t, alpha, aux)

                xpm_0 += alpha * A_0   # Also adds to xp_0 as xpm_0 is a view
                x_new[:,:nOdes] += alpha * A_t
                # for i in range(1,q+1):
                #     x_new[:,nOdes*i:(nOdes*(i+1))] = alpha * A_t

                x0_twice = 2*xp_0
                # x_new[-1] = xp_0
                # xpm_0 += alpha * A_t[-1]   # Also adds to xp_0 as xpm_0 is a view
                # x_new[:,:nOdes] += alpha * A_t
                x_new[-1,:] = xp_0[:]

                # y_pert = make_y_pert(x_new.T, paramGuess, self.tolerance)
                # A_j0 = np.vstack((np.zeros(nOdes), y_pert*np.eye(nOdes, dtype=np.float64)))
                # xp_0 = np.hstack((x_0, A_j0.flat)) # Add perturbed ICs as extra states
                # xp_0[nOdes:] = np.tile(xp_0[:nOdes], (q+1,)) + A_j0.flat  # Add perturbations to ICs
                # xpm_0 = np.reshape(xp_0, (q+2, nOdes))  # Matrix 'view' of xp_0
                # x0_twice = 2*x_new[-1,:]
                ctr += 1

            x_guess = x_new
            err0 = err1

        # Return initial guess if it failed to converge
        sol = solinit

        if converged:
            if len(tspan) > 2:
                _, x_out = mcpi(self.mcpi_eom, tspan, x_new[-1, :nOdes], args=(const,))
            else:
                x_out = np.flipud(x_new)[:,:nOdes]
                tspan = t_arr[::-1]

            sol.x, sol.y = tspan, x_out[:,:nXandLam].T
            sol.parameters = x_out[0,nXandLam:]
            sol.arcs = [(0, len(sol.x)-1)]
            sol.arc_seq = (0,)
            sol.extra = x_new

        sol.converged = converged
        sol.aux = aux
        return sol



import scipy.integrate
import logging


def ode45(f, tspan, y0, *args, **kwargs):
    """Implements interface similar to MATLAB's ode45 using scipy"""

    abstol = kwargs.get('abstol', 1e-5)
    reltol = kwargs.get('reltol', 1e-5)
    maxstep = kwargs.get('maxstep', 0.1)

    sol = scipy.integrate.solve_ivp(lambda t, y: f(t, y, *args), tspan, y0, rtol=reltol, atol=abstol, max_step=maxstep)

    return sol.t, sol.y


def warning(err, string):
    logging.warning('warning: ' + err)
    logging.warning(string)


def error(err, string):
    logging.error('error: ' + err)
    logging.error(string)
    raise RuntimeError(string)  # Raise error to notify shooting solver
    # exit

# specify default arguments here:


def processOdeArgs(**kwargs):
    defaults = {'normcontrol': 'on',  # ('on','off')
                # index array describing components that should be nonnegative
                'nonnegative': None,
                # relative tolerance (can be array of same dimension as y0)
                'reltol': 1e-5,
                'abstol': 1e-5,			# absolute tolerance
                'outputsel': None,		# which components to save
                # save output every outputsave steps
                'outputsave': None,
                'initialstep': None,
                'maxstep': None,
                'mass': None,
                'stats': 'off'			# statistics
                }

    if len(kwargs) > 0:
        for i in kwargs:
            defaults[i.lower()] = kwargs[i]

    # accept True instead of 'on' for certain options
    opts = ['stats', 'normcontrol']
    for o in opts:
        if defaults[o]:
            defaults[o] = 'on'

    return defaults



# Source : http://vjethava.blogspot.com/2010/11/matlabs-keyboard-command-in-python.html
import sys
import code

def keyboard(banner=None):
    ''' Function that mimics the matlab keyboard command '''
    # use exception trick to pick up the current frame
    try:
        raise None
    except:
        frame = sys.exc_info()[2].tb_frame.f_back
    print("# Use quit() to exit :) Happy debugging!")
    # evaluate commands in current namespace
    namespace = frame.f_globals.copy()
    namespace.update(frame.f_locals)
    try:
        code.interact(banner=banner, local=namespace)
    except SystemExit:
        return



"""
Computes the necessary conditions of optimality using Bryson & Ho's method

[1] Bryson, Arthur Earl. Applied optimal control: optimization, estimation and control. CRC Press, 1975.
"""

from .optimlib import *
from beluga.utils import sympify
import simplepipe as sp


def make_control_law(dhdu, controls):
    """
    Solves control equation to get control law.

    :param dhdu:
    :param controls:
    :return: Control law options.
    """
    try:
        logging.info(controls)
        var_list = list(controls)
        logging.info("Attempting using SymPy ...")
        logging.debug("dHdu = "+str(dhdu))
        ctrl_sol = sympy.solve(dhdu, var_list, dict=True, minimal=True, simplify=False)

        # raise ValueError() # Force mathematica
    except ValueError as e:  # FIXME: Use right exception name here
        logging.debug(e)
        logging.info("No control law found")
        from beluga.utils_old.pythematica import mathematica_solve
        logging.info("Attempting using Mathematica ...")
        var_sol = mathematica_solve(dhdu,var_list)
        # TODO: Extend numerical control laws to mu's
        ctrl_sol = var_sol
    logging.info('Control found')
    logging.info(ctrl_sol)
    control_options = ctrl_sol
    return control_options


def generate_problem_data(workspace):
    """Generates the `problem_data` dictionary used for code generation."""

    tf_var = sympify('tf')
    problem_data = {
        'method': 'brysonho',
        'problem_name': workspace['problem_name'],
        'aux_list': [
            {
            'type' : 'const',
            'vars': [str(k) for k in workspace['constants']]
            }
        ],
        'state_list':
            [str(x) for x in it.chain(workspace['states'], workspace['costates'])]
        ,
        'parameter_list': [str(tf_var)] + [str(p) for p in workspace['parameters']],
        'x_deriv_list': [str(tf_var*state.eom) for state in workspace['states']],
        'lam_deriv_list':[str(tf_var*costate.eom) for costate in workspace['costates']],
        'deriv_list':
            [str(tf_var*state.eom) for state in workspace['states']] +
            [str(tf_var*costate.eom) for costate in workspace['costates']]
        ,
        'states': workspace['states'],
        'costates': workspace['costates'],
        'constants': workspace['constants'],
        'parameters': [tf_var] + workspace['parameters'],
        'controls': workspace['controls'],
        'mu_vars': workspace['mu_vars'],
        'quantity_vars': workspace['quantity_vars'],

        'costate_eoms': workspace['costate_eoms'],
        'bc_list': workspace['bc_list'],
        'ham': workspace['ham'],
        's_list': workspace['s_list'],
        'num_states': 2*len(workspace['states']),
        'num_params': len(workspace['parameters']) + 1,
        'dHdu': workspace['dhdu'],
        'bc_initial': [str(_) for _ in workspace['bc_initial']],
        'bc_terminal': [str(_) for _ in workspace['bc_terminal']],
        'control_options': workspace['control_law'],
        'control_list': [str(u) for u in workspace['controls']+workspace['mu_vars']],
        'num_controls': len(workspace['controls'])+len(workspace['mu_vars']),
        'ham_expr': str(workspace['ham']),
        'quantity_list': workspace['quantity_list'],
        'nOdes': 2*len(workspace['states'])
    }

    return problem_data


# Implement workflow using simplepipe and functions defined above
BrysonHo = sp.Workflow()
BrysonHo.add_task(BaseWorkflow)
BrysonHo.add_task(make_control_law, inputs=('dhdu','controls'), outputs=('control_law'))
BrysonHo.add_task(generate_problem_data, inputs='*', outputs=('problem_data'))
traditional = BrysonHo



#!/usr/bin/env python
"""Beluga Optimal Control Solver.

Usage:
  beluga --config
  beluga (-v | --version)
  beluga (-h | --help)
  beluga SCENARIO
                  [-o | --output <file>]
                  ([--nolog] | [-l | --log <level>])
                  ([-q] | [-d | --display <level>])

Options:
  -h, --help                show this screen and exit
  -v, --version             show version

  Logging options
  -l, --log <level>         specify minimum logging level [default: ERROR]
  -l0,-l1,-l2,-l3,-l4       shortcuts for ALL, INFO, WARN, ERROR and CRITICAL
                            respectively
  -loff, --nolog            suppress logging, equivalent to --log=off

  -q                        quiet mode, equivalent to --display=off
  -d, --display <level>     specify minimum verbose output level [default: INFO]
  -d0,-d1,-d2,-d3,-d4       shortcuts for ALL, INFO, WARN, ERROR and CRITICAL
                            respectively

  -o, --output <file>       specify data file for solution
                            (overrides option specified in input file)

Arguments:
  SCENARIO                  name of python module orpath to python/json/yaml file
                            containing the problem scenario [REQUIRED]

  <file>                    path to the data file for solution
  <level>                   logging level [ALL, INFO, WARN, ERROR, CRITICAL, OFF]

Example:
  Run problem using python module name :
    beluga brachisto

  Run using path to input file:
    beluga /path/to/problem.py

  Specify logging and display levels (0 -> all messages including debug messages)
    beluga -d0 -l0 /path/to/brachisto.py

  Specify output data file
    beluga brachisto --output=brachisto_out.dill

  Run configuration tool
    beluga --config

Help:
  For help using this tool, please open an issue on the Github repository:
  https://github.rcac.purdue.edu/RDSL/beluga

"""
import os
import sys
import logging
import importlib

import docopt

def load_scenario(scenario_name):
    """Loads a scenario from python module name or file name/path"""
    # TODO: Log error messages on failure

    # Check if a python filename was given
    if scenario_name.endswith('.py') and os.path.exists(scenario_name) and os.path.isfile(scenario_name):
        module_dir, module_file = os.path.split(scenario_name)
        module_name, module_ext = os.path.splitext(module_file)
        sys.path.append(module_dir)
    elif (scenario_name.endswith('.yml') or scenario_name.endswith('.json'))and os.path.exists(scenario_name) and os.path.isfile(scenario_name):
        print('Loading from YAML scenario ..')
        return load_yaml(scenario_name)
    else:
        if scenario_name.isidentifier():
            sys.path.append(os.getcwd())
            module_name = scenario_name
        else:
            print('Invalid scenario filename or module name')
            return None
    try:
        scenario = importlib.import_module(module_name)
         # Check if module has a get_problem() function
        # if hasattr(scenario,'get_problem') and callable(scenario.get_problem):
        #     # Module loaded successfully
        #     # print('Module loaded successfully. 😂')
        #     return scenario.get_problem()
        # else:
        #     print('Unable to find get_problem function in scenario module')
        #     return None

    except ImportError as e:
        print('Error loading scenario: ')
        import traceback
        traceback.print_exc(e)
        return None


def main():
    options = docopt.docopt(__doc__,version=0.1)

    levels = {  'ALL': logging.DEBUG,
                'DEBUG': logging.DEBUG,
                '0': logging.DEBUG,
                'INFO': logging.INFO,
                '1': logging.INFO,
                'WARNING': logging.WARN,
                'WARN': logging.WARN,
                '2': logging.WARN,
                'ERROR': logging.ERROR,
                '3': logging.ERROR,
                'CRITICAL': logging.CRITICAL,
                '4': logging.CRITICAL,
                'OFF': logging.CRITICAL + 1}

    # Process logging options
    if options['--nolog']:
        # Suppress all logging
        options['--log'][0] = 'off'

    if options['--log'][0].upper() not in levels:
        print('Invalid value specified for logging level')
        return
    logging_lvl = levels[options['--log'][0].upper()]

    # Process console output options
    if options['-q']:
        # Suppress all console output
        options['--display'][0] = 'off'

    if options['--display'][0].upper() not in levels:
        print('Invalid value specified for display level')
        return
    display_lvl = levels[options['--display'][0].upper()]

    if len(options['--output']) > 0:
        output = os.path.abspath(options['--output'][0].strip())
        # Check if the file locaton is writeable
        if not os.access(os.path.dirname(output), os.W_OK):
            print('Unable to access output file location or invalid filename 😭 😭')
            return
    else:
        output = None


    import builtins
    # import beluga
    beluga = sys.modules['beluga']
    builtins.beluga = beluga

    beluga.setup_beluga(logging_level=logging_lvl, display_level=display_lvl, output_file=output)

    scenario = load_scenario(options['SCENARIO'].strip())
    if scenario is None:
        return



"""
Base functions required by all optimization methods.
"""


from beluga.utils import sympify, sympify2, keyboard
from beluga.problem import SymVar
import sympy
import functools as ft
import itertools as it
import simplepipe as sp
import re as _re
from beluga.problem import SymVar
import logging
import numpy as np
import sympy as sym


def add_equality_constraints(hamiltonian, equality_constraints):
    """
    Adjoins equality constraints to a Hamiltonian function.

    .. math::
        \\begin{aligned}
            \\text{add_equality_constraints} : C^\\infty(M) &\\rightarrow C^\\infty(M) \\\\
            (H, f) &\\mapsto H + \\mu_i f_i \\; \\forall \\; i \\in f
        \\end{aligned}

    :param ham: A Hamiltonian function, :math:`H`.
    :param equality_constraints: List of equality constraints, :math:`f`.
    :return: Augmented Hamiltonian function.
    """
    # Adjoin equality constraints
    for i in range(len(equality_constraints)):
        hamiltonian += sympify('mu'+str(i+1)) * (sympify2(equality_constraints[i].expr))

    return hamiltonian


def get_satfn(var, ubound=None, lbound=None, slopeAtZero=1):
    """
    Documentation needed.

    :param var:
    :param ubound:
    :param lbound:
    :param slopeAtZero:
    :return:
    """

    # var -> variable inside saturation function
    if ubound is None and lbound is None:
        raise ValueError('At least one bound should be specified for the constraint.')
    if ubound == lbound:
        raise ValueError('The upper and lower bounds should be different.')

    # Symmetric bounds
    if ubound is None:
        # Return lower bound sat fn
        #ubound = -lbound
        satfcn = lbound + sym.exp(var)
        print('Using one-sided saturation function', satfcn)
        return satfcn
    elif lbound is None:
        # Return upper bound sat fn
        satfcn = ubound - sym.exp(-var)
        print('Using one-sided saturation function', satfcn)
        return satfcn
    else:
        s = 4*slopeAtZero/(ubound - lbound)
        print('Using two sided saturation function')
        return ubound - (ubound - lbound)/(1 + sym.exp(s*var))


def init_workspace(ocp, guess):
    """
    Initializes the simplepipe workspace using an OCP definition.

    All the strings in the original definition are converted into symbolic
    expressions for computation.

    :param ocp: An optimal control problem.
    :return:
    """

    workspace = {}
    # variable_list = ['states', 'controls', 'constraints', 'quantities', 'initial_cost', 'terminal_cost', 'path_cost']
    workspace['problem_name'] = ocp.name
    workspace['indep_var'] = SymVar(ocp._properties['independent'])
    workspace['states'] = [SymVar(s) for s in ocp.states()]
    workspace['controls'] = [SymVar(u) for u in ocp.controls()]
    workspace['constants'] = [SymVar(k) for k in ocp.constants()]

    constraints = ocp.constraints()
    workspace['constraints'] = {c_type: [SymVar(c_obj, sym_key='expr') for c_obj in c_list]
                                for c_type, c_list in constraints.items()
                                if c_type != 'path'}

    workspace['constraints_adjoined'] = ocp.constraints().adjoined

    workspace['path_constraints'] = [SymVar(c_obj, sym_key='expr', excluded=('direction', 'start_eps'))
                                     for c_obj in constraints.get('path', [])]

    workspace['quantities'] = [SymVar(q) for q in ocp.quantities()]
    workspace['initial_cost'] = SymVar(ocp.get_cost('initial'), sym_key='expr')
    workspace['terminal_cost'] = SymVar(ocp.get_cost('terminal'), sym_key='expr')
    workspace['path_cost'] = SymVar(ocp.get_cost('path'), sym_key='expr')
    guess.dae_num_states = 0
    return workspace


# TODO: Check if this function is ever used. I don't think it's needed.
def jacobian(expr_list, var_list, derivative_fn):
    """
    Returns a Jacobian matrix for a given set of functions and variables.

    .. math::
        \\begin{aligned}
            \\text{jacobian} : \\Gamma^1(M) &\\rightarrow \\Gamma^2(M) \\\\
            (f, x, d) &\\mapsto J_{ij} = d_{x_j} f_i \\; \\forall \\; i,j
        \\end{aligned}

    :param expr_list: List of expressions to take the partials of, :math:`f`.
    :param var_list: List of variables to take the partials with respect to, :math:`x`.
    :param derivative_fn: Derivative function, :math:`d`
    :return: The Jacobian matrix, :math:`J_{ij}`.
    """

    jac = sympy.zeros(len(expr_list), len(var_list))
    for i, expr in enumerate(expr_list):
        for j, var in enumerate(var_list):
            jac[i, j] = derivative_fn(expr, var)
    return jac


def make_augmented_cost(cost, constraints, constraints_adjoined, location):
    """
    Augments the cost function with the given list of constraints.

    .. math::
        \\begin{aligned}
            \\text{make_augmented_cost} : C^\\infty(M) &\\rightarrow C^\\infty(M) \\\\
            (f, g) &\\mapsto f + g_i \\nu_i \\; \\forall \\; i \\in g
        \\end{aligned}

    :param cost: The cost function, :math:`f`.
    :param constraints: List of constraint to adjoin to the cost function, :math:`g`.
    :param constraints_adjoined: Boolean value on whether or not the adjoined method is used. Skips if `False`.
    :param location: Location of each constraint.

    Returns the augmented cost function
    """

    if not constraints_adjoined:
        return cost

    lagrange_mult = make_augmented_params(constraints, constraints_adjoined, location)
    aug_cost_expr = cost.expr + sum(nu * c for (nu, c) in zip(lagrange_mult, constraints[location]))

    aug_cost = SymVar({'expr':aug_cost_expr, 'unit': cost.unit}, sym_key='expr')
    return aug_cost


def make_augmented_params(constraints, constraints_adjoined, location):
    """
    Make the lagrange multiplier terms for adjoining boundary conditions.

    :param constraints: List of constraints at the boundaries.
    :param constraints_adjoined: Boolean value on whether or not the adjoined method is used.
    :param location: Location of each constraint.
    :return: Lagrange multipliers for the given constraints.
    """
    if not constraints_adjoined:
        return []

    def make_lagrange_mult(c, ind = 1):
        return sympify('lagrange_' + location + '_' + str(ind))

    lagrange_mult = [make_lagrange_mult(c, ind) for (ind,c) in enumerate(constraints[location], 1)]

    return lagrange_mult


def make_bc_mask(states, controls, mu_vars, cost, aug_cost, derivative_fn):

    """Creates mask marking free and bound variables at t=0

    free = 1, constrained = 0
    """
    aug_cost_expr = (aug_cost.expr - cost.expr)

    state_mask = []
    costate_mask = []
    for state in states:
        if derivative_fn(aug_cost_expr, state) == 0:
            # State not constrained -> costate = 0
            state_mask.append(1)    # State free
            costate_mask.append(0)  # Costate fixed
        else:
            state_mask.append(0)    # State constrained
            costate_mask.append(1)  # Costate free

    return state_mask+costate_mask+[1]*(len(controls)+len(mu_vars)+1)  # Add tf as free param


def make_boundary_conditions(constraints, constraints_adjoined, states, costates, cost, derivative_fn, location,
                             prefix_map=(('initial',(r'([\w\d\_]+)_0', r"_x0['\1']", sympify('-1'))),
                                         ('terminal',(r'([\w\d\_]+)_f', r"_xf['\1']", sympify('1'))))):
    """
    Creates boundary conditions for initial and terminal constraints.

    :param constraints: List of boundary constraints.
    :param states: List of state variables.
    :param costates: List of costate variables.
    :param cost: Cost function.
    :param derivative_fn: Total derivative function.
    :param location: Location of each boundary constraint.
    :param prefix_map: Prefix mapping.
    :return: List of boundary conditions.
    """
    prefix_map = dict(prefix_map)
    bc_list = [sanitize_constraint_expr(x, states, location, prefix_map) for x in constraints[location]]
    *_, sign = dict(prefix_map)[location]

    cost_expr = sign * cost

    #TODO: Fix hardcoded if conditions
    #TODO: Change to symbolic
    if constraints_adjoined:
        bc_list += [str(costate - derivative_fn(cost_expr, state)) for state, costate in zip(states, costates)]
    else:
        refd = [sum(derivative_fn(c, s) for c in constraints[location]) for s in states]
        bc_list += [str(costate - derivative_fn(cost_expr, state)) for i, (state, costate) in enumerate(zip(states,costates)) if refd[i] == False]

    return bc_list


def make_constrained_arc_fns(workspace):
    """
    Creates constrained arc control functions.

    :param workspace:
    :return:
    """
    controls = workspace['controls']
    constants = workspace['constants']
    states = workspace['states']
    costates = workspace['costates']
    parameters = workspace['parameters']
    quantity_vars = workspace['quantity_vars']
    fn_args_lamdot = [list(it.chain(states, costates)), parameters, constants, controls]
    # control_fns = [workspace['control_fn']]
    tf_var = sympify('tf')
    costate_eoms = [ {'eom':[str(_.eom*tf_var) for _ in workspace['costates']], 'arctype':0} ]
    bc_list = [] # Unconstrained arc placeholder

    mu_vars = workspace['mu_vars']


    for arc_type, s in enumerate(workspace['s_list'],1):
        pi_list = [str(_) for _ in s['pi_list']]


        costate_eom = {'eom':[str(_.eom*tf_var) for _ in s['lamdot']],
                       'arctype':arc_type,
                       'pi_list': pi_list}

        entry_bc, exit_bc = make_constraint_bc(s,
                                workspace['states'],
                                workspace['costates'],
                                workspace['parameters'],
                                workspace['constants'],
                                workspace['controls'], mu_vars, workspace['quantity_vars'], workspace['ham'])
        bc = {'entry_bc': entry_bc,
              'exit_bc': exit_bc,
              'arctype': arc_type,
              'pi_list': pi_list,
              'name': s['name']}
        bc_list.append(bc)

        # control_fns.append(u_fn)
        costate_eoms.append(costate_eom)

    # yield control_fns
    yield costate_eoms
    yield bc_list


def make_constraint_bc(s, states, costates, parameters, constants, controls, mu_vars, quantity_vars, ham):
    """
    Documentation needed.

    :param s:
    :param states:
    :param costates:
    :param parameters:
    :param constants:
    :param controls:
    :param mu_vars:
    :param quantity_vars:
    :param ham:
    :return:
    """
    num_states = len(states)
    costate_slice = slice(num_states, 2*num_states)
    ham_aug = s['ham']
    corner_conditions = s['corner']
    tangency = s['tangency']
    tf_var = sympify('tf')

    y1m = sympy.symbols(' '.join('_'+str(_.name)+'_1m' for _ in it.chain(states, costates, [tf_var])))
    y1m_x = sympy.Matrix([y1m[:num_states]])
    y1m_l = sympy.Matrix([y1m[costate_slice]])

    y1p = sympy.symbols(' '.join('_'+str(_.name)+'_1p' for _ in it.chain(states, costates, [tf_var])))
    y1p_x = sympy.Matrix([y1p[:num_states]])
    y1p_l = sympy.Matrix([y1p[costate_slice]])

    y2m = sympy.symbols(' '.join('_'+str(_.name)+'_2m' for _ in it.chain(states, costates, [tf_var])))
    y2p = sympy.symbols(' '.join('_'+str(_.name)+'_2p' for _ in it.chain(states, costates, [tf_var])))
    y2m = sympy.Matrix([y2m])
    y2p = sympy.Matrix([y2p])

    u_m = sympy.symbols(' '.join('_'+str(_.name)+'_m' for _ in it.chain(controls, mu_vars)))
    u_p = sympy.symbols(' '.join('_'+str(_.name)+'_p' for _ in it.chain(controls, mu_vars)))
    if not hasattr(u_m, '__len__'):
        u_m = (u_m,)
    if not hasattr(u_p, '__len__'):
        u_m = (u_p,)
    def make_subs(in_vars, out_vars):
        return {k: v for k,v in zip(in_vars, out_vars)}

    subs_1m = make_subs(it.chain(states, costates, [tf_var], controls, mu_vars), it.chain(y1m, u_m))
    subs_1p = make_subs(it.chain(states, costates, [tf_var], controls, mu_vars), it.chain(y1p, u_p))
    subs_2m = make_subs(it.chain(states, costates, [tf_var], controls, mu_vars), it.chain(y2m, u_m))
    subs_2p = make_subs(it.chain(states, costates, [tf_var], controls, mu_vars), it.chain(y2p, u_p))

    ham1m = ham.subs(quantity_vars).subs(subs_1m)
    ham1p = ham_aug.subs(quantity_vars).subs(subs_1p)
    ham2m = ham_aug.subs(quantity_vars).subs(subs_2m)
    ham2p = ham.subs(quantity_vars).subs(subs_2p)
    tangency_1m = sympy.Matrix(tangency).subs(subs_1p)

    entry_bc = [
        *tangency_1m,  # Tangency conditions, N(x,t) = 0
        *(y1m_x - y1p_x), # Continuity in states at entry
        *(y1m_l - y1p_l - corner_conditions.subs(subs_1p)), # Corner condns on costates
        ham1m - ham1p
    ]
    exit_bc = [
        *(y2m - y2p)[:-1], # Continuity in states and costates at exit (excluding tf)
        ham2m - ham2p
    ]
    # bc_arc = [*tangency_1m,  # Tangency conditions, N(x,t) = 0
    #           *(y1m_x - y1p_x), # Continuity in states at entry
    #           *(y1m_l - y1p_l - corner_conditions), # Corner condns on costates
    #
    #           ham1m - ham1p,
    #           ham2m - ham2p]

    entry_bc = [str(_) for _ in entry_bc]
    exit_bc = [str(_) for _ in exit_bc]
    return entry_bc, exit_bc


def make_costate_names(states):
    """
    Makes a list of variables representing each costate.

    :param states: List of state variables, :math:`x`.
    :return: List of costate variables, :math:`\\lambda_x`.
    """

    return [sympify('lam'+str(s.name).upper()) for s in states]


def make_costate_rates(ham, states, costate_names, derivative_fn):
    """
    Makes a list of rates of change for each of the costates.

    :param ham: Hamiltonian function.
    :param states: List of state variables.
    :param costate_names: List of costate variables.
    :param derivative_fn: Total derivative function.
    :return: Rates of change for each costate.
    """

    costates = [SymVar({'name': lam, 'eom':derivative_fn(-1*(ham), s)}) for s, lam in zip(states, costate_names)]
    return costates


#TODO: Determine if make_dhdu() is ever even used. Like 2 of the functions show up as not imported.
def make_dhdu(ham, controls, derivative_fn):
    """
    Computes the partial of the hamiltonian w.r.t control variables.

    :param ham: Hamiltonian function.
    :param controls: A list of each control variable.
    :param derivative_fn: Total derivative function.
    :return: :math:`dH/du`
    """

    dhdu = []
    for ctrl in controls:
        dHdu = derivative_fn(ham, ctrl)
        custom_diff = dHdu.atoms(sympy.Derivative)
        # Substitute "Derivative" with complex step derivative
        repl = {(d, im(f.func(v+1j*1e-30))/1e-30) for d in custom_diff
                    for f,v in zip(d.atoms(sympy.AppliedUndef), d.atoms(Symbol))}

        dhdu.append(dHdu.subs(repl))

    return dhdu


def make_ham_lamdot_with_eq_constraint(states, constraints, path_cost, derivative_fn):
    """
    Creates a Hamiltonian function and costate rates.

    :param states: A list of state variables, :math:`x`.
    :param constraints: A set of constraints.
    :param path_cost: The path cost to be minimized.
    :param derivative_fn: The derivative function.
    :return: A Hamiltonian function, :math:`H`.
    :return: A list of costate rates, :math:`\\dot{\\lambda}_x`
    """

    costate_names = make_costate_names(states)
    ham = path_cost.expr + sum([lam*s.eom for s, lam in zip(states, costate_names)])
    ham = add_equality_constraints(ham, constraints.get('equality', []))
    yield ham
    yield make_costate_rates(ham, states, costate_names, derivative_fn)


def make_parameters(initial_lm_params, terminal_lm_params, s_list):
    """
    Makes parameters.

    :param initial_lm_params:
    :param terminal_lm_params:
    :param s_list:
    :return: List of parameters.
    """
    all_pi_names = [p['pi_list'] for p in s_list]
    params_list = [str(p) for p in it.chain(initial_lm_params,
                                            terminal_lm_params)] #, *all_pi_names)]
    if len(params_list) > 0:
        parameters = [sympy.symbols(p) for p in params_list]
    else:
        parameters = []

    return parameters


def make_time_bc(constraints, bc_terminal):
    """
    Makes free or fixed final time boundary conditions.

    :param constraints: List of constraints.
    :param bc_terminal: Terminal boundary condition.
    :return: New terminal boundary condition.
    """
    time_constraints = constraints.get('independent', [])
    if len(time_constraints) > 0:
        return bc_terminal+['tf - 1']
    else:
        # Add free final time boundary condition
        return bc_terminal+['_H - 0']


def process_path_constraints(workspace):
    """
    Documentation needed.

    :param workspace:
    :return:
    """
    states = workspace['states']
    controls = workspace['controls']
    constants = workspace['constants']
    constraints = workspace['constraints']
    derivative_fn = workspace['derivative_fn']
    quantity_vars = workspace['quantity_vars']
    quantity_list = workspace['quantity_list']

    path_cost_expr = workspace['path_cost'].expr
    path_cost_unit = workspace['path_cost'].unit
    if path_cost_expr == 0:
        logging.debug('No path cost specified, using unit from terminal cost function')
        path_cost_expr = None
        path_cost_unit = workspace['terminal_cost'].unit

    logging.debug('Path cost is of unit: '+str(path_cost_unit))
    time_unit = workspace['indep_var'].unit
    path_constraints = workspace['path_constraints']

    eq = constraints.get('equality', [])
    xi_init_vals = []
    for (ind, c) in enumerate(path_constraints):
        # Determine order of constraint
        logging.debug('Processing path constraint: '+str(c.name))
        order = 0
        cq = [c.expr]
        dxdt = [state.eom for state in states]

        # Zeroth order constraints have no 'xi' state
        xi_vars = []
        h = []
        while True:
            control_found = False
            for u in controls:
                if u in cq[-1].subs(quantity_vars).atoms():
                    logging.info('Constraint is of order '+str(order))
                    control_found = True
                    break
            if control_found:
                break

            dcdx = [derivative_fn(cq[-1], state) for state in states]

            # Chain rule (Assume there is no explciit time-dependence) to find dcdt
            cq.append(sum(d1*d2 for d1,d2 in zip(dcdx, dxdt)))
            order = order + 1

            # Create the auxiliary state variables
            xi_vars.append(sym.Symbol('xi'+str(ind+1)+str(order)))

        # Create the smoothing control variable
        xi_vars.append(sym.Symbol('ue'+str(ind+1)))

        # TODO: Fix constraint object to accept two limits
        c_limit = sym.sympify(c.bound)
        if c_limit.is_Number:
            # TODO: Allow continuation on constraints
            # Define new hidden constant
            c_limit = sympify('_'+str(c.name))
            constants.append(SymVar({'name': str(c_limit), 'value': float(c.bound), 'unit': c.unit}))
            logging.debug('Added constant '+str(c_limit))

        c_lbound = c_ubound = None
        if c.direction == '>':
            c_lbound = c_limit
        elif c.direction == '<':
            c_ubound = c_limit
        elif c.direction == '<>':
            c_ubound = c_limit
            c_lbound = -c_limit
        else:
            raise ValueError('Invalid direction specified for constraint')

        psi = get_satfn(xi_vars[0], ubound=c_ubound, lbound=c_lbound, slopeAtZero=1)
        psi_vars = [(sympify2('psi'+str(ind+1)+'0('+str(xi_vars[0])+')'), psi)]

        # Add to quantity list
        quantity_vars['psi'+str(ind+1)+'0'] = str(psi)
        quantity_list.append({'name':('psi'+str(ind+1)+'0'), 'expr':str(psi)})

        # m-th order constraint needs up to m-th derivative of psi to be defined
        psi_i = psi
        psi_vars_deriv = []
        psi_var_func = [(psi_vars[0][0], sym.Symbol('psi'+str(ind+1)+'0'))]
        for i in range(order):

            psi_i = derivative_fn(psi_i, xi_vars[0])
            # psi_vars.append((Symbol('psi'+str(ind+1)+str(i+1)+'('+str(xi_vars[0])+')'), psi_i))
            current_psi_var = sym.Symbol('psi'+str(ind+1)+str(i+1))
            current_psi_var_func = sympify2(str(current_psi_var)+'('+str(xi_vars[0])+')')
            psi_vars.append((current_psi_var, psi_i))
            psi_var_func.append((current_psi_var_func, current_psi_var))
            psi_vars_deriv.append((sympify2('Derivative(psi'+str(ind+1)+str(i)+'('+str(xi_vars[0])+'), '+str(xi_vars[0])+')'),
                                   current_psi_var_func))
            quantity_vars[str(current_psi_var)] = str(psi_i)
            quantity_list.append({'name':str(current_psi_var), 'expr':str(psi_i)})

        # psi_vars = psi_vars + []
        # psi_var_sub = [(v,k) for k,v in psi_vars]

        # FIXME: Hardcoded h derivatives for now
        # h = [psi_vars[0][0]]
        # h.append(psi_vars[1][0]*xi_vars[1]) # psi'*xi12
        # h.append(psi_vars[2][0]*xi_vars[1] + psi_vars[1][0]*xi_vars[2]) # psi''*xi12 + psi'*xi13
        # psi'''*xi12 + xi13*psi12'' + psi12*xi13 + psi11*ue1
        # h.append(psi_vars[3][0]*xi_vars[1] + 2 * psi_vars[2][0]*xi_vars[2] + psi_vars[1][0]*xi_vars[3] )

        #TODO: Hardcoded 't' as independent variable with unit of 's'
        # c_vals = [80e3, -5000, 9.539074102210087] # third number is vdot at zero approx
        c_vals = np.ones(order)*0.1
        h = [psi_vars[0][0]]

        for i in range(order):
            # Add 'xi' state
            states.append(SymVar({'name':str(xi_vars[i]), 'eom':str(xi_vars[i+1]), 'unit':(c.unit/(time_unit**i))}))
            # Constraint all cq at initial point (forms constraints for xi_ij)
            # ocp.constraints().initial(str(cq[i] - h[i]),'('+c.unit+')/s^('+str(i)+')')
            constraints['terminal'].append(SymVar({'expr':str(cq[i] - h[i].subs(psi_var_func)), 'unit':c.unit/(time_unit**i)}, sym_key='expr'))

            # Add to initial guess vector
            print(f'Adding {xi_vars[i]} = {c_vals[i]}')
            xi_init_vals.append(c_vals[i])

            dhdxi = [derivative_fn(h[i], xi_v).subs(psi_vars_deriv) for xi_v in xi_vars[:-1]]
            dhdt  = sum(d1*d2 for d1,d2 in zip(dhdxi,xi_vars[1:])) # xi11dot = xi12 etc.
            # dhdt = dhdt.subs(psi_var_sub)
            h.append(dhdt)

        h = [h_i.subs(psi_var_func) for h_i in h]
        # Add the smoothing control with the right unit
        ue_unit = c.unit/(time_unit**order)
        controls.append(SymVar({'name':str(xi_vars[-1]), 'unit':str(ue_unit)}))

        logging.debug('Adding control '+str(xi_vars[-1])+' with unit '+str(ue_unit))

        # Add equality constraint
        cqi_unit = ue_unit*time_unit
        # ocp.constraints().equality(str(cq[-1] - h[-1]),str(cqi_unit))

        eq.append(SymVar({'type':'equality', 'expr':str(cq[-1] - h[-1]), 'unit':str(cqi_unit), 'name':'_eq_'+str(ind)}))

        # Add smoothing factor
        eps_const = sympify('eps_'+str(c.name))
        eps_unit = (path_cost_unit/ue_unit**2)/time_unit #Unit of integrand
        eps_unit = sym.Symbol('nd')
        if not c.start_eps:
            c.start_eps = 1
        constants.append(SymVar({'name':eps_const, 'value': c.start_eps, 'unit':str(eps_unit)}))
        logging.debug('Adding smoothing factor '+str(eps_const)+' with unit '+str(eps_unit))

        # Append new control to path cost
        if path_cost_expr is None:
            path_cost_expr = eps_const*xi_vars[-1]**2
        else:
            path_cost_expr = path_cost_expr + eps_const*xi_vars[-1]**2

    constraints['equality'] = eq
    if path_cost_expr is not None:
        logging.debug('Updated path cost is: '+str(path_cost_expr))
        path_cost = SymVar({'expr': path_cost_expr, 'unit':path_cost_unit}, sym_key='expr')
    else:
        path_cost = SymVar({'expr': '0', 'unit':path_cost_unit}, sym_key='expr')
    derivative_fn = ft.partial(total_derivative, dependent_vars=quantity_vars)
    jacobian_fn = ft.partial(jacobian, derivative_fn=derivative_fn)

    s_list = []
    mu_vars = []

    yield states
    yield controls
    yield constants
    yield constraints
    yield path_cost
    yield s_list
    yield mu_vars
    yield xi_init_vals
    yield derivative_fn
    yield jacobian_fn


def process_quantities(quantities):
    """
    Performs preprocessing on quantity definitions. Creates a new total
    derivative operator that takes considers these definitions.

    :param quantities: List of quantities.
    :return:
    """

    # TODO: Sanitize quantity expressions
    # TODO: Check for circular references in quantity expressions

    # Trivial case when no quantities are defined
    if len(quantities) == 0:
        yield {}
        yield []
        yield total_derivative
        yield ft.partial(jacobian, derivative_fn=total_derivative)

    quantity_subs = [(q.name, q.value) for q in quantities]
    quantity_sym, quantity_expr = zip(*quantity_subs)
    quantity_expr = [qty_expr.subs(quantity_subs) for qty_expr in quantity_expr]

    # Use substituted expressions to recreate quantity expressions
    quantity_subs = [(str(qty_var),qty_expr) for qty_var, qty_expr in zip(quantity_sym, quantity_expr)]
    # Dictionary for substitution
    quantity_vars = dict(quantity_subs)

    # Dictionary for use with mustache templating library
    quantity_list = [{'name':str(qty_var), 'expr':str(qty_expr)} for qty_var, qty_expr in zip(quantity_sym, quantity_expr)]

    # Function partial that takes derivative while considering quantities
    derivative_fn = ft.partial(total_derivative, dependent_vars=quantity_vars)
    jacobian_fn = ft.partial(jacobian, derivative_fn=derivative_fn)

    yield quantity_vars
    yield quantity_list
    yield derivative_fn
    yield jacobian_fn


def sanitize_constraint_expr(constraint, states, location, prefix_map):
    """
    Checks the initial/terminal constraint expression for invalid symbols
    Also updates the constraint expression to reflect what would be in code

    :param constraint: List of constraints.
    :param states: List of state variables.
    :param location: Location of the constraints.
    :param prefix_map: Prefix mapping.
    """

    if location not in prefix_map:
        raise ValueError('Invalid constraint type')

    pattern, prefix, _ = dict(prefix_map)[location]
    m = _re.findall(pattern,str(constraint.expr))
    invalid = [x for x in m if x not in states]

    if not all(x is None for x in invalid):
        raise ValueError('Invalid expression(s) in boundary constraint:\n'+str([x for x in invalid if x is not None]))

    return _re.sub(pattern,prefix,str(constraint.expr))


def total_derivative(expr, var, dependent_vars=None):
    """
    Take derivative taking pre-defined quantities into consideration

    :param expr: Expression to evaluate the derivative of.
    :param var: Variable to take the derivative with respect to.
    :param dependent_vars: Other dependent variables to consider with chain rule.
    """
    if dependent_vars is None:
        dependent_vars = {}

    dep_var_names = dependent_vars.keys()
    dep_var_expr = [(expr) for (_,expr) in dependent_vars.items()]

    dFdq = [sympy.diff(expr, dep_var).subs(dependent_vars.items()) for dep_var in dep_var_names]
    dqdx = [sympy.diff(qexpr, var) for qexpr in dep_var_expr]

    # Chain rule + total derivative
    out = sum(d1*d2 for d1,d2 in zip(dFdq, dqdx)) + sympy.diff(expr, var)
    return out


BaseWorkflow = sp.Workflow([
    sp.Task(init_workspace, inputs=('problem','guess'), outputs='*'),
    sp.Task(process_quantities,
            inputs=('quantities'),
            outputs=('quantity_vars', 'quantity_list', 'derivative_fn', 'jacobian_fn')),
    sp.Task(process_path_constraints, inputs='*',
            outputs=('states', 'controls', 'constants', 'constraints', 'path_cost', 's_list', 'mu_vars', 'xi_init_vals', 'derivative_fn', 'jacobian_fn')),
    sp.Task(ft.partial(make_augmented_cost, location='initial'),
            inputs=('initial_cost', 'constraints', 'constraints_adjoined'),
            outputs=('aug_initial_cost')),
    sp.Task(ft.partial(make_augmented_params, location='initial'),
            inputs=('constraints', 'constraints_adjoined'),
            outputs=('initial_lm_params')),
    sp.Task(ft.partial(make_augmented_cost, location='terminal'),
            inputs=('terminal_cost', 'constraints', 'constraints_adjoined'),
            outputs=('aug_terminal_cost')),
    sp.Task(ft.partial(make_augmented_params, location='terminal'),
            inputs=('constraints', 'constraints_adjoined'),
            outputs=('terminal_lm_params')),
    # sp.Task(make_costate_names,
    #         inputs=('states'),
    #         outputs=('costate_names')),
    sp.Task(make_ham_lamdot_with_eq_constraint,
            inputs=('states', 'constraints', 'path_cost', 'derivative_fn'),
            outputs=('ham', 'costates')),
    sp.Task(ft.partial(make_boundary_conditions, location='initial'),
            inputs=('constraints', 'constraints_adjoined', 'states', 'costates', 'aug_initial_cost', 'derivative_fn'),
            outputs=('bc_initial')),
    sp.Task(ft.partial(make_boundary_conditions, location='terminal'),
            inputs=('constraints', 'constraints_adjoined', 'states', 'costates', 'aug_terminal_cost', 'derivative_fn'),
            outputs=('bc_terminal')),
    sp.Task(make_time_bc, inputs=('constraints', 'bc_terminal'), outputs=('bc_terminal')),
    sp.Task(make_dhdu,
            inputs=('ham', 'controls', 'derivative_fn'),
            outputs=('dhdu')),
    sp.Task(make_parameters, inputs=['initial_lm_params', 'terminal_lm_params', 's_list'],
            outputs='parameters'),
    sp.Task(make_bc_mask,
            inputs=('states', 'controls', 'mu_vars', 'initial_cost', 'aug_initial_cost', 'derivative_fn'),
            outputs=('bc_free_mask')),
    sp.Task(make_constrained_arc_fns, inputs='*', outputs=['costate_eoms', 'bc_list'])])



from .Scaling import Scaling
from .Solution import Solution
# from .bvpinit import bvpinit


# __all__ = ['Algorithm','Solution','BVP','FunctionTemplate','bvpinit']
import os
import glob
modules = glob.glob(os.path.dirname(__file__)+"/*.py")
__all__ = [ os.path.basename(f)[:-3] for f in modules]



from . import renderers, datasources
from .elements import PlotList,Plot
from .renderers import BaseRenderer
from .datasources import BaseDataSource
import dill, inspect, logging
import os
class BelugaPlot:
    """
    Manages the plotting framework
    """
    def __init__(self, filename = None, datasource='dill', renderer = 'matplotlib', default_step = -1, default_sol = -1):
        """
        Initializes plotting framework with given data file
        """
        self._plots = []
        self._figures = []
        self._plotconfig = {}
        self.global_settings = {}
        self.filename = filename
        self.default_step_idx = default_step
        self.default_sol_idx = default_sol

        # Load datasource by filename unless one is specified directly
        if filename is not None:
            fname, file_ext = os.path.splitext(filename)
        else:
            file_ext = None

        # TODO: Get default datasource information from global configuration
        if isinstance(datasource, BaseDataSource):
            # Use custom renderer object
            self.datasource = datasource
        else:
            # Load renderer from the list of existing classes
            self.datasource = None
            for name, obj in inspect.getmembers(datasources):
                if inspect.isclass(obj) and issubclass(obj, BaseDataSource):
                    if (isinstance(datasource, str) and name.lower() == datasource.lower()) \
                        or file_ext in obj.valid_exts:
                        # Renderer initialized with its default settings
                        self.datasource = obj(filename)
            if self.datasource is None:
                raise ValueError('Datasource "'+datasource+'" not found')

        # TODO: Get default renderer information from global configuration
        if isinstance(renderer, BaseRenderer):
            # Use custom renderer object
            self.renderer = renderer
        else:
            # Load renderer from the list of existing classes
            self.renderer = None
            for name, obj in inspect.getmembers(renderers):
                if inspect.isclass(obj) and issubclass(obj, BaseRenderer):
                    if name.lower() == renderer.lower():
                        # Renderer initialized with its default settings
                        self.renderer = obj()
            if self.renderer is None:
                raise ValueError('Renderer "'+renderer+'" not found')

    def add_plot(self, step = None, sol = None, datasource = None, colormap=None, mesh_size=None):
        """
        Adds a new plot
            (alias for add_plot() in PlotList)
        """
        if step is None:
            step = self.default_step_idx
        if sol is None:
            sol = self.default_sol_idx
        if datasource is None:
            datasource = self.datasource

        plot = Plot(step, sol, mesh_size, datasource, colormap)
        self._plots.append(plot)
        return plot

    def render(self,show=True):
        for plot in self._plots:
            plot.preprocess()
            fig = self.renderer.create_figure()
            self.renderer.render_plot(fig, plot)
            self._figures.append(fig)
            for fn in plot.postprocess_list:
                fn(self.renderer, fig, plot)
        if show:
            self.show()

    def show(self):
        self.renderer.show_all()

    def plotconfig(varname,value=None):
        """
        Adds/updates settings affecting all plots
        """
        if isinstance(varname, dict):
            self._plotconfig.update(varname)
        else:
            self._plotconfig[varname] = value
        return self



import numbers as num

from sympy import *
from beluga.utils import sympify, keyboard

import copy

class Scaling(dict):
    excluded_aux = ['function']

    def __init__(self):
        self.units = {}
        self.units_sym = []
        self.scale_func = {}
        self.problem_data = {}

    """Defines scaling for a set of units"""
    def unit(self, unit_str, unit_scale):
        """Adds scaling factor for a given unit
        Allows method chaining
        """
        self.units[unit_str] = unit_scale
        return self

    def initialize(self, ws):
        """Initializes the scaling process.

        Parameters
        ----------
        ws - dict
            Workspace processed by OCP workflow
        """
        self.problem_data = ws['problem_data']

        # Generate scaling functions for states, costates
        # constants, constraints, lagrange multipliers

        # Convert units to symbols
        self.units_sym = symbols(list(self.units))

        # Growing list TODO: Put inside utils
        # TODO: Automate the following sections

        # Scaling functions for constants
        self.scale_func['const'] = {str(const): self.create_scale_fn(const.unit)
                                    for const in ws['constants']}

        # Cost function used for scaling costates
        cost_keys = ['path_cost', 'terminal_cost', 'initial_cost']
        cost_used = next(key for key in cost_keys if str(ws[key]) != '0')
        if len(cost_used) < 1:
            raise ValueError('At least one cost function must be specified as nonzero!')
        cost_unit = str(ws[cost_used].unit)

        # Scaling functions for states & costates (combined)
        self.scale_func['states'] = {}
        self.scale_func['states'] = {str(state): self.create_scale_fn(state.unit)
                                     for state in ws['states']}
        costate_units = {str(costate): self.create_scale_fn('('+cost_unit+')/('+str(state.unit)+')')
                         for state, costate in zip(ws['states'],ws['costates']) }
        self.scale_func['states'].update(costate_units)

        self.scale_func['initial'] = self.scale_func['states']
        self.scale_func['terminal'] = self.scale_func['states']

        # Scaling functions for constraint multipliers and other parameters
        self.scale_func['parameters'] = {}

        # Scaling function for the independent variable
        # TODO: Fix hardcoding
        self.scale_func['parameters']['tf'] = self.create_scale_fn(ws['indep_var'].unit)

        self.scale_func['constraints'] = {}

        for c_type, c_list in ws['constraints'].items():
            for c, mul_var in zip(c_list, ws.get(c_type+'_lm_params',[])):
                mul_unit = '('+cost_unit+')/('+str(c.unit)+')'
                self.scale_func['parameters'][str(mul_var)] = self.create_scale_fn(mul_unit)

        # Scale factors for path constraint-related stuff
        for s in ws['s_list']:
            for pi_var in s['pi_list']:
                pi_unit = '('+cost_unit+')/('+str(s['unit'])+')'
                self.scale_func['parameters'][str(pi_var)] = self.create_scale_fn(pi_unit)
            self.scale_func['constraints'][str(s['name'])] = self.create_scale_fn(s['unit'])

    def create_scale_fn(self,unit_expr):
        return lambdify(self.units_sym,sympify(unit_expr))

    def compute_base_scaling(self,sol,scale_expr):
        if isinstance(scale_expr,num.Number):
            # If scaling factor is a number, use it
            return scale_expr
        else:
            variables = [(aux_name,aux_val)
                    for aux_type in sol.aux
                    if isinstance(sol.aux[aux_type],dict)
                    for (aux_name,aux_val) in sol.aux[aux_type].items()
                    ]
            # Have to do in this order to override state values with arrays
            variables += [(state,max(abs(sol.y[idx,:])))
                            for idx,state in enumerate(self.problem_data['state_list'])]

            var_dict = dict(variables)

            # Evaluate expression to get scaling factor
            return float(sympify(scale_expr).subs(var_dict,dtype=float).evalf())

    def compute_scaling(self, sol):
        from collections import OrderedDict
        # Units should be stored in order to be used as function arguments
        self.scale_factors = OrderedDict()
        # Evaluate scaling factors for each base unit
        for (unit,scale_expr) in self.units.items():
            self.scale_factors[unit] = self.compute_base_scaling(sol, scale_expr)

        # Ordered list of unit scaling factors for use as function parameters
        scale_factor_list = [v for (k,v) in self.scale_factors.items()]

        # Find scaling factors for each entity in problem
        self.scale_vals = {}

        for var_type,var_funcs in self.scale_func.items():
            # If there are no sub items, use the scale factor directly
            if callable(var_funcs):
                self.scale_vals[var_type] = var_funcs(*scale_factor_list)
            else:
                # Else call scaling function for each sub item
                self.scale_vals[var_type] = {}
                for var_name,var_func in var_funcs.items():
                    self.scale_vals[var_type][var_name] = var_func(*scale_factor_list)

    def scale(self, sol):
        """Scales a BVP solution"""
        solout = copy.deepcopy(sol)

        # Additional aux entries for initial and terminal BCs
        extras = [{'type':'initial','vars':self.problem_data['state_list']},
                  {'type':'terminal','vars':self.problem_data['state_list']}]

        # Scale the states and costates
        for idx, state in enumerate(self.problem_data['state_list']):
            solout.y[idx, :] /= self.scale_vals['states'][state]

        # Scale auxiliary variables
        for aux in (self.problem_data['aux_list']+extras):
            if aux['type'] not in Scaling.excluded_aux:
                for var in aux['vars']:
                    solout.aux[aux['type']][var] /= self.scale_vals[aux['type']][var]

        # Scale parameters
        for idx, param in enumerate(self.problem_data['parameter_list']):
            solout.parameters[idx] /= self.scale_vals['parameters'][param]

        for (s_name, arc_idx), s_val in sol.aux['constraint'].items():
            scale_val = self.scale_vals['constraints'][s_name]
            solout.aux['constraint'][(s_name, arc_idx)] /= scale_val

        return solout

    def unscale(self, sol):
        """ Unscales a solution object"""
        solout = copy.deepcopy(sol)

        # Additional aux entries for initial and terminal BCs
        extras = [{'type': 'initial', 'vars': self.problem_data['state_list']},
                  {'type': 'terminal', 'vars': self.problem_data['state_list']}]

        # Scale the states and costates
        for idx, state in enumerate(self.problem_data['state_list']):
            solout.y[idx, :] *= self.scale_vals['states'][state]

        # Scale auxiliary variables
        for aux in (self.problem_data['aux_list']+extras):
            if aux['type'] not in Scaling.excluded_aux:
                for var in aux['vars']:
                    solout.aux[aux['type']][var] *= self.scale_vals[aux['type']][var]

        # Scale parameters
        for idx, param in enumerate(self.problem_data['parameter_list']):
            solout.parameters[idx] *= self.scale_vals['parameters'][param]

        for (s_name, arc_idx), s_val in solout.aux['constraint'].items():
            scale_val = self.scale_vals['constraints'][s_name]
            solout.aux['constraint'][(s_name, arc_idx)] *= scale_val

        return solout



from beluga.bvpsol.algorithms.BaseAlgorithm import BaseAlgorithm
import numpy as np
import copy
from scipy.optimize import minimize
import logging
from beluga.utils import keyboard

class Collocation(BaseAlgorithm):
    """
    Collocation algorithm for solving boundary value problems.
    """
    def __new__(cls, *args, **kwargs):
        """
        Creates a new Collocation object.

        :param args: Unused
        :param kwargs: Additional parameters accepted by the solver.
        :return: Collocation object.

        +------------------------+-----------------+-----------------+
        | Valid kwargs           | Default Value   | Valid Values    |
        +========================+=================+=================+
        | cached                 | True            | Bool            |
        +------------------------+-----------------+-----------------+
        | tolerance              | 1e-4            | > 0             |
        +------------------------+-----------------+-----------------+
        | max_error              | 100             | > 0             |
        +------------------------+-----------------+-----------------+
        | max_iterations         | 100             | > 0             |
        +------------------------+-----------------+-----------------+
        | number_of_nodes        | 30              | > 1             |
        +------------------------+-----------------+-----------------+
        | use_numba              | False           | Bool            |
        +------------------------+-----------------+-----------------+
        | verbose                | False           | Bool            |
        +------------------------+-----------------+-----------------+
        """
        
        obj = super(Collocation, cls).__new__(cls, *args, **kwargs)

        cached = kwargs.get('cached', True)
        tolerance = kwargs.get('tolerance', 1e-4)
        max_error = kwargs.get('max_error', 100)
        max_iterations = kwargs.get('max_iterations', 100)
        number_of_nodes = kwargs.get('number_of_nodes', 30)
        use_numba = kwargs.get('use_numba', False)
        verbose = kwargs.get('verbose', False)

        obj.cached = cached
        obj.tolerance = tolerance
        obj.max_error = max_error
        obj.max_iterations = max_iterations
        obj.number_of_nodes = number_of_nodes
        obj.use_numba = use_numba
        obj.verbose = verbose
        return obj
    
    def solve(self, deriv_func, quad_func, bc_func, solinit):
        """
        Solve a two-point boundary value problem using the collocation method.

        :param deriv_func: The ODE function.
        :param quad_func: The quad func.
        :param bc_func: The boundary conditions function.
        :param solinit: An initial guess for a solution to the BVP.
        :return: A solution to the BVP.
        """
        sol = copy.deepcopy(solinit)
        sol.y = sol.y.T
        sol.set_interpolate_function('cubic')
        number_of_datapoints = len(sol.t)
        if number_of_datapoints < 3:
            raise ValueError('Initial guess must have more than 2 data points in collocation.')

        reconstruct = False
        if self.number_of_nodes != number_of_datapoints:
            new_t = np.linspace(sol.t[0], sol.t[-1], self.number_of_nodes)
            (new_y, new_q, new_u) = sol(new_t)
            sol.t = new_t
            sol.y = new_y
            sol.q = new_q
            sol.u = new_u
            if quad_func is not None:
                raise NotImplemented # TODO: Put reconstruction of q's in Trajectory()? Or leave in ivpsol?

        # Set up some required class functions for collocation
        self.eoms = deriv_func
        self.quadratures = quad_func
        self.path_cost = None # TODO: Assumed to be indirect. This class can handle direct with some work.
        self.terminal_cost = None # TODO: Assumed to be indirect.
        self.bcs = bc_func

        # self.constraint = {'type': 'eq', 'fun': self._collocation_constraint}
        self.constraint_midpoint = {'type': 'eq', 'fun': self._collocation_constraint_midpoint}
        self.constraint_boundary = {'type': 'eq', 'fun': self._collocation_constraint_boundary}

        # Set up initial guess and other info
        self.tspan = sol.t
        self.number_of_odes = len(sol.y[0])

        # TODO: The following if-then structure is silly, but I can't resolve this until some optimlib corrections are made
        if sol.q is None:
            self.number_of_quads = 0
            sol.q = np.array([], dtype=np.float64)
        elif len(sol.q) == 0:
            self.number_of_quads = 0
            sol.q = np.array([], dtype=np.float64)
        else:
            self.number_of_quads = len(sol.q)
            raise NotImplementedError

        if sol.parameters is None:
            self.number_of_params = 0
            sol.parameters = np.array([], dtype=np.float64)
        else:
            self.number_of_params = len(sol.parameters)

        vectorized = self._wrap_params(sol.y, sol.q, sol.parameters)

        self.aux = sol.aux
        sol.converged = False

        if self.verbose:
            logging.info('Running collocation... ')

        xopt = minimize(self._collocation_cost, vectorized, args=(), method='SLSQP', jac=None, hess=None, hessp=None, bounds=None, constraints=[self.constraint_midpoint, self.constraint_boundary], tol=self.tolerance, callback=None, options=None)

        if self.verbose:
            logging.info(xopt['message'])

        if xopt['status'] == 0:
            sol.converged = True

        # Organize the output with the sol() structure
        sol.t = self.tspan
        sol.y, sol.q, sol.parameters = self._unwrap_params(xopt['x'])
        sol.y = sol.y.T
        return sol

    def reconstruct(self, time, ivp):
        # TODO: Rewrite this trash. Use ivpsol(), this should already be done.
        if time < ivp.sol.x[0] or time > ivp.sol.x[-1]:
            raise ValueError

        if time == ivp.sol.x[0]:
            return ivp.sol.x[0], ivp.sol.y[0], ivp.sol.quads

        ind = 1
        t_high = ivp.sol.x[ind]
        tf = ivp.sol.x[-1]
        quads = ivp.sol.quads
        while t_high < time:
            if ivp.quadratures is not None:
                t0 = ivp.sol.x[ind-1]
                t1 = ivp.sol.x[ind]
                t12 = (t1 + t0) / 2
                p0 = ivp.sol.y[ind-1]
                p1 = ivp.sol.y[ind]
                try:
                    dp0 = np.squeeze(ivp.eoms(t0, p0, ivp.sol.params, ivp.sol.consts))
                except:
                    keyboard()
                dp1 = np.squeeze(ivp.eoms(t1, p1, ivp.sol.params, ivp.sol.consts))
                midpoint_predicted = 1 / 2 * (p0 + p1) + tf / (len(ivp.sol.x) - 1) / 8 * (dp0 - dp1)
                dt = ivp.sol.x[ind] - ivp.sol.x[ind-1]
                dquads = np.squeeze(ivp.quadratures(t0, p0, ivp.sol.params, ivp.sol.consts)) * dt
                dquads12 = np.squeeze(ivp.quadratures(t12, midpoint_predicted, ivp.sol.params, ivp.sol.consts)) * dt
                dquads1 = np.squeeze(ivp.quadratures(t1, p1, ivp.sol.params, ivp.sol.consts)) * dt
                C1, C2, C3, C4 = self._get_poly_coefficients_1_3(quads, dquads, dquads12, dquads1)
                quads = C1 + C2 + C3 + C4
            ind += 1
            t_high = ivp.sol.x[ind]

        dt = ivp.sol.x[ind] - ivp.sol.x[ind - 1]
        t_temp = (time - ivp.sol.x[ind - 1]) / dt
        p0 = np.array(ivp.sol.y[ind - 1])
        dp0 = np.squeeze(ivp.eoms(ivp.sol.x[ind - 1], ivp.sol.y[ind - 1], ivp.sol.params, ivp.sol.consts)) * dt
        p1 = np.array(ivp.sol.y[ind])
        dp1 = np.squeeze(ivp.eoms(ivp.sol.x[ind], ivp.sol.y[ind], ivp.sol.params, ivp.sol.consts)) * dt
        if ivp.quadratures is not None:
            t0 = ivp.sol.x[ind - 1]
            t1 = ivp.sol.x[ind]
            t12 = (t1 + t0) / 2
            midpoint_predicted = 1 / 2 * (p0 + p1) + tf / (len(ivp.sol.x) - 1) / 8 * (dp0 - dp1)
            dquads = np.squeeze(ivp.quadratures(t0, p0, ivp.sol.params, ivp.sol.consts)) * dt
            dquads12 = np.squeeze(ivp.quadratures(t12, midpoint_predicted, ivp.sol.params, ivp.sol.consts)) * dt
            dquads1 = np.squeeze(ivp.quadratures(t1, p1, ivp.sol.params, ivp.sol.consts)) * dt
            C1, C2, C3, C4 = self._get_poly_coefficients_1_3(quads, dquads, dquads12, dquads1)
            quadsf = C1 + C2*t_temp + C3*t_temp**2 + C4*t_temp**3
        else:
            quadsf = quads

        C1, C2, C3, C4 = self._get_poly_coefficients_2_2(p0, dp0, p1, dp1)
        y_new = C1 + C2*t_temp + C3*t_temp**2 + C4*t_temp**3
        return time, y_new, quadsf

    def _collocation_constraint_midpoint(self, vectorized):
        y, quads0, params = self._unwrap_params(vectorized)
        tf = self.tspan[-1]
        # dX = np.squeeze(self.eoms(self.tspan, X.T, params, self.aux)).T # TODO: Vectorized our code compiler so this line works
        dX = np.squeeze([self.eoms(ti, yi, params, self.aux) for ti,yi in zip(self.tspan, y)])
        dp0 = dX[:-1]
        dp1 = dX[1:]
        p0 = y[:-1]
        p1 = y[1:]
        t0 = self.tspan[:-1]
        t1 = self.tspan[1:]
        t12 = (t0+t1)/2
        midpoint_predicted = 1 / 2 * (p0 + p1) + tf / (self.number_of_nodes - 1) / 8 * (dp0 - dp1)
        midpoint_derivative_predicted = -3 / 2 * (self.number_of_nodes - 1) / tf * (p0 - p1) - 1 / 4 * (dp0 + dp1)
        midpoint_derivative_actual = np.squeeze([self.eoms(ti, yi, params, self.aux) for ti, yi in zip(t12, midpoint_predicted)]) # TODO: Vectorize, so this one works as well
        outvec = midpoint_derivative_predicted - midpoint_derivative_actual
        return outvec.flatten()

    def _collocation_constraint_boundary(self, vectorized):
        X, quads0, params = self._unwrap_params(vectorized)
        if len(quads0) == 0:
            return np.squeeze(self.bcs(np.array([X[0]]).T, np.array([X[-1]]).T, params, self.aux))
        else:
            quadsf = self._integrate_quads(self.tspan, X, quads0, params, self.aux, quads=self.quadratures)
            return np.squeeze(self.bcs(self.tspan[0], X[0], self.tspan[-1], X[-1], quads0, quadsf, params, self.aux))

    def _collocation_cost(self, vectorized):
        X, quads0, params = self._unwrap_params(vectorized)
        return 0

        if self.path_cost is not None:
            path_cost = self.path_cost(self.tspan, X, params, self.consts)
            if path_cost is None:
                path_cost = 0
            else:
                path_cost = np.trapz(path_cost, x=self.tspan)
        else:
            path_cost = 0

        if self.terminal_cost is not None:
            terminal_cost = self.terminal_cost(self.tspan[-1], X[-1], params, self.consts)
            if terminal_cost is None:
                terminal_cost = 0
        else:
            terminal_cost = 0
        return path_cost + terminal_cost

    def _unwrap_params(self, vectorized):
        # TODO: This is hella inefficient
        if self.number_of_params + self.number_of_quads == 0:
            X = vectorized.reshape([self.number_of_nodes,self.number_of_odes])
            quads = np.array([])
            params = np.array([])
        else:
            quads = np.array([])
            params = np.array([])
            X = vectorized[0:-(self.number_of_params+self.number_of_quads)].reshape([self.number_of_nodes,self.number_of_odes])
            if self.number_of_params == 0:
                quads = vectorized[-self.number_of_quads:]
            elif self.number_of_quads == 0:
                params = np.array(vectorized[-self.number_of_params:])
            else:
                quads = vectorized[-(self.number_of_params+self.number_of_quads):-(self.number_of_params)]
                params = vectorized[-self.number_of_params:]

        return X, quads, params

    def _wrap_params(self, y, q, params):
        return np.concatenate((y.flatten(), q, params))

    @staticmethod
    def _get_poly_coefficients_1_3(p0, dquads0, dquads12, dquads1):
        C1 = p0
        C2 = dquads0
        C3 = -3/2*dquads0 + 2*dquads12 - 1/2*dquads1
        C4 = 2/3*dquads0 - 4/3*dquads12 + 2/3*dquads1
        return C1, C2, C3, C4

    @staticmethod
    def _get_poly_coefficients_2_2(p0, dp0, p1, dp1):
        C1 = p0
        C2 = dp0
        C3 = -3*p0 - 2*dp0 + 3*p1 - dp1
        C4 = 2*p0 + dp0 - 2*p1 + dp1
        return C1, C2, C3, C4

    @staticmethod
    def _collocation_midpoint_prediction(p0, dp0, p1, dp1, tf, N):
        return 1 / 2 * (p0 + p1) + tf / (N - 1) / 8 * (dp0 - dp1)

    @staticmethod
    def _collocation_midpoint_derivative(p0, dp0, p1, dp1, tf, N):
        return -3 / 2 * (N - 1) / tf * (p0 - p1) - 1 / 4 * (dp0 + dp1)

    @classmethod
    def _get_default_options(cls, options='default'):
        """ Default options structure for Collocation. """
        if options == 'default':
            return [1, 10]
        elif options == 'quiet':
            return [0, 10]


"""
problem2 -- Rename to 'problem' after refactoring.

Contains class/functions related to defining the optimal control problems.
"""

import scipy.optimize
import numpy as np
import dill

import json
import logging
import os.path
import re
from functools import partialmethod
from collections import namedtuple, ChainMap
from itertools import zip_longest

from beluga.bvpsol import Scaling
from beluga.ivpsol.integrators import ode45
from beluga.utils import sympify, tic, toc
from beluga.ivpsol import Propagator

Cost = namedtuple('Cost', ['expr', 'unit'])
class OCP(object):
    """Builder class for defining optimal control problem."""

    def __init__(self, name=''):
        """Initializes problem object.

        Parameters
        ----------
        name - str
            Unique name for the problem
        """
        self.name = self._format_name(name)

        self._systems = {'default': {}}  # Dynamic system definitions
        self._properties = {}  # Problem properties
        self._constraints = ConstraintList()

        self._scaling = Scaling()

    # Alias for returning cost function by type
    def get_cost(self, cost_type):
        """Retrieves the cost function for the problem.

        Parameters
        ----------
        cost_type - str
            Type of cost function - path, initial or terminal
        """
        return self._properties.get(cost_type + '_cost', {'expr':'0','unit':'0'})

    def set_cost(self, expr, unit, cost_type):
        """Sets cost function for problem.

        Parameters
        ----------
        expr - str
            Expression for cost function

        unit - str
            Unit of cost function

        cost_type - str
            Type of cost function - path, initial or terminal
        """
        self._properties[cost_type+'_cost'] = {'expr':expr, 'unit':unit}

    def set_property(self, *args, property_name, property_args, **kwargs):
        """
        Adds a property of the optimal control problem

        Parameters
        ----------
        args

        property_name

        property_args

        kwargs

        Returns a reference to self for chaining purposes
        """
        prop = self._properties.get(property_name, [])
        prop.append(_combine_args_kwargs(property_args, args, kwargs))
        self._properties[property_name] = prop
        return self

    def get_property(self, property_name):
        """
        Returns the property specified by the name
        """
        return self._properties.get(property_name, [])

    # TODO: Write documentation for these aliases
    state = partialmethod(set_property, property_name='states',
                    property_args=('name', 'eom', 'unit'))
    control = partialmethod(set_property, property_name='controls',
                    property_args=('name', 'unit'))
    constant = partialmethod(set_property, property_name='constants',
                    property_args=('name', 'value', 'unit'))
    quantity = partialmethod(set_property, property_name='quantities',
                    property_args=('name', 'value'))

    states = partialmethod(get_property, property_name='states')
    controls = partialmethod(get_property, property_name='controls')
    constants = partialmethod(get_property, property_name='constants')
    quantities = partialmethod(get_property, property_name='quantities')

    # TODO: Maybe write as separate function?
    def independent(self, name, unit):
        self._properties['independent'] = {'name': name, 'unit':unit}

    # Aliases for defining properties of the problem
    path_cost = partialmethod(set_cost, cost_type='path')
    initial_cost = partialmethod(set_cost, cost_type='initial')
    terminal_cost = partialmethod(set_cost, cost_type='terminal')

    Lagrange = path_cost
    Mayer = terminal_cost


    def constraints(self):
        """
        Returns the ConstraintList object containing alias methods

        This function is purely for aesthetic purposes while method chaining
        in the input file
        """
        return self._constraints

    def scale(self, **scale_mappings):
        """Defines scaling for dimensional units in the problem."""
        for unit, scale_expr in scale_mappings.items():
            self._scaling.unit(unit, scale_expr)

    def __str__(self):
        """
        Returns a string representation of the object
        """
        return str({'name': self.name,
                    'properties': self._properties,
                    'constraints': self._constraints,
                    'continuation': str(self.continuation)})

    def _format_name(self, name):
        """Validates that the name is in the right format
            Only alphabets, numbers and underscores allowed
            Should not start with a number or underscore

            Required for the in-memory compilation of code to work
        """

        if re.match(r'[a-zA-Z]\w+', name):
            return name
        else:
            raise ValueError("""Invalid problem name specified.
            Only alphabets, numbers and underscores allowed
            Should start with an alphabet""")

    def as_json(self):
        """Converts the problem definition into a pure dictionary."""
        output = self._properties
        output['problem_name'] = self.name
        output['constraints'] = self._constraints
        return json.dumps(output)


class ConstraintList(dict):
    def __new__(cls, *args, **kwargs):
        obj = super(ConstraintList, cls).__new__(cls, *args, **kwargs)
        obj.adjoined = False
        return obj

    def set_adjoined(self, bool):
        self.adjoined = bool

    def add_constraint(self, *args, constraint_type='', constraint_args=[], **kwargs):
        """
        Adds constraint of the specified type

        Returns reference to self.constraint_aliases for chaining
        """

        c_list = self.get(constraint_type, [])

        constraint = _combine_args_kwargs(constraint_args, args, kwargs)
        c_list.append(constraint)
        self[constraint_type] = c_list

        return self

    # Aliases for defining constraints of different types
    constraint_args = ('expr', 'unit')
    initial = partialmethod(add_constraint, constraint_type='initial',
                constraint_args=constraint_args)
    terminal = partialmethod(add_constraint, constraint_type='terminal',
                constraint_args=constraint_args)
    equality = partialmethod(add_constraint, constraint_type='equality',
                constraint_args=constraint_args)
    interior_point = partialmethod(add_constraint, constraint_type='interior_point',
                constraint_args=constraint_args)
    independent = partialmethod(add_constraint, constraint_type='independent',
                constraint_args=constraint_args)
    path = partialmethod(add_constraint, constraint_type='path',
                constraint_args=('name', 'expr', 'direction', 'bound', 'unit', 'start_eps')
                )

    # def get(self, constraint_type):
    #     """
    #     Returns list of constraints of a specific type
    #     """
    #     return [c for c in self if c.type == constraint_type]

def _combine_args_kwargs(arg_list, args, kwargs, fillvalue=''):
    """Combines positional and keyword arguments

    Parameters
    ----------
    arg_list - list of str
        List of keys in order of positional arguments

    args - list of str
        List of positional arguments

    kwargs: dict
        Dictionary of keyword arguments

    Returns
    -------
    A dictionary merging kwargs and args with keys from
    from args_list

    Example
    -------
    >>> _combine_args_kwargs(['foo','bar'],[1,2],{'baz':3})
    {'foo':1, 'bar':2, 'baz': 3}
    """
    pos_args = {key: val for (key, val) in
                zip_longest(arg_list, args, fillvalue=fillvalue)}
    arg_dict = dict(ChainMap(kwargs, pos_args))
    return (arg_dict)


class SymVar(object):
    """
    Represents an object that can be used in SymPy and is created from a dict
    """

    def __init__(self, param_dict, sym_key='name', excluded=()):
        self.__dict__ = {k: sympify(v) if k not in excluded else v
                         for k, v in param_dict.items()}
        self.param_list = list(param_dict.keys())
        if sym_key is not None:
            self._sym = self.__dict__[sym_key]
        else:
            self._sym = None

    def _sympy_(self):
        """
        Makes the object usable in sympy expressions directly
        """
        return self._sym

    def __hash__(self):
        return hash(self._sym)

    def __eq__(self, other):
        return self._sym == other._sym

    def __repr__(self):
        return str(self._sym)

    def keys(self):
        return self.param_list

    def __getitem__(self, key):
        return getattr(self, key)

    def __eq__(self, other):
        return str(self._sym) == str(other)

BVP = namedtuple('BVP', 'deriv_func bc_func compute_control path_constraints')
BVP.__new__.__defaults__ = (None,) # path constraints optional

class GuessGenerator(object):
    """Generates the initial guess from a variety of sources."""

    def __init__(self, **kwargs):
        self.setup_funcs = {'auto': self.setup_auto,
                            'file': self.setup_file,
                            'static': self.setup_static,
                            }
        self.generate_funcs = {'auto': self.auto,
                               'file': self.file,
                               'static': self.static
                               }
        self.setup(**kwargs)
        self.dae_num_states = 0

    def setup(self, mode='auto', **kwargs):
        """Sets up the initial guess generation process"""

        self.mode = mode
        if mode in self.setup_funcs:
            self.setup_funcs[mode](**kwargs)
        else:
            raise ValueError('Invalid initial guess mode specified')

        return self

    def generate(self, *args):
        """Generates initial guess data from given settings"""
        if self.mode in self.generate_funcs:
            return self.generate_funcs[self.mode](*args)
        else:
            raise ValueError('Invalid initial guess mode specified')

    def setup_static(self, solinit=None):
        self.solinit = solinit

    def static(self, bvp_fn, solinit):
        """Directly specify initial guess structure"""
        return self.solinit

    def setup_file(self, filename='', step=0, iteration=0):
        self.filename = filename
        self.step = step
        self.iteration = iteration
        if not os.path.exists(self.filename) or not os.path.isfile(self.filename):
            logging.error('Data file ' + self.filename + ' not found.')
            raise ValueError('Data file not found!')

    def file(self, bvp_fn, solinit):
        """Generates initial guess by loading an existing data file.

        bvp_fn : BVP
            BVP object containing functions

        solinit : Solution
            Solution object with some starting information (such as aux vars)
        """
        logging.info('Loading initial guess from ' + self.filename)
        fp = open(self.filename, 'rb')
        out = dill.load(fp)
        if self.step >= len(out['solution']):
            logging.error('Continuation step index exceeds bounds. Only ' +
                          str(len(out['solution']))
                          + ' continuation steps found.')
            raise ValueError('Initial guess step index out of bounds')

        if self.iteration >= len(out['solution'][self.step]):
            logging.error('Continuation iteration index exceeds bounds. Only '
                          + str(len(out['solution'][self.step]))
                          + ' iterations found.')
            raise ValueError('Initial guess iteration index out of bounds')

        sol = out['solution'][self.step][self.iteration]
        # sol.extra = None

        fp.close()

        logging.info('Initial guess loaded')
        return sol

    def setup_auto(self, start=None,
                   direction='forward',
                   time_integrate=0.1,
                   costate_guess=0.1,
                   control_guess=0.1,
                   use_control_guess=False,
                   param_guess=None):
        """Setup automatic initial guess generation"""

        if direction in ['forward', 'reverse']:
            self.direction = direction
        else:
            raise ValueError('Direction must be either forward or reverse.')

        self.time_integrate = abs(time_integrate)
        if time_integrate == 0:
            raise ValueError('Integration time must be non-zero')

        # TODO: Check size against number of states here
        self.start = start
        self.costate_guess = costate_guess
        self.param_guess = param_guess
        self.control_guess = control_guess
        self.use_control_guess = use_control_guess

    def auto(self, bvp_fn, solinit, param_guess=None):
        """Generates initial guess by forward/reverse integration."""

        # Assume normalized time from 0 to 1
        tspan = [0, 1]

        x0 = np.array(self.start)

        # Add costates
        if isinstance(self.costate_guess, float):
            x0 = np.r_[x0, self.costate_guess * np.ones(len(self.start))]
        else:
            x0 = np.r_[x0, self.costate_guess]

        if isinstance(self.control_guess, float):
            u0 = self.control_guess*np.ones(self.dae_num_states)
        else:
            u0 = self.control_guess

        # Add time of integration to states

        # x0 = np.append(x0, self.time_integrate)

        # Guess zeros for missing parameters
        # TODO: Automatically generate parameter guess values

        if param_guess is None:
            param_guess = np.zeros(len(solinit.aux['parameters']))
        elif len(param_guess) < len(solinit.aux['parameters']):
            param_guess += np.zeros(len(solinit.aux['parameters']) - len(param_guess))
        elif len(param_guess) > len(solinit.aux['parameters']):
            # TODO: Write a better error message
            raise ValueError('param_guess too big. Maximum length allowed is ' +
                             len(solinit.aux['parameters']))

        param_guess[0] = self.time_integrate

        if self.dae_num_states > 0:
            dae_guess = u0
            if not self.use_control_guess:
                dhdu_fn = bvp_fn.get_dhdu_func(0, x0, param_guess, solinit.aux)

                dae_x0 = scipy.optimize.fsolve(dhdu_fn, dae_guess, xtol=1e-5)
            else:
                dae_x0 = dae_guess

            print('dae_x0',dae_x0)
            x0 = np.append(x0, dae_x0)  # Add dae states

        print('guess ode',id(bvp_fn.deriv_func))
        logging.debug('Generating initial guess by propagating: ')
        logging.debug(str(x0))

        if self.direction == 'reverse':
            tspan = [0, -1]

        tic()
        prop = Propagator()
        solivp = prop(bvp_fn.deriv_func_ode45, None, tspan, x0, [], param_guess, solinit.aux)
        elapsed_time = toc()
        logging.debug('Propagated initial guess in %.2f seconds' % elapsed_time)
        solinit.t = solivp.t
        solinit.y = solivp.y.T # TODO: This shouldn't be transposed, but it's gonna require significant reworking.
        solinit.parameters = param_guess
        return solinit



from beluga.ivpsol import Propagator, Trajectory, integrate_quads
import numpy as np
from math import *

tol = 1e-3

def test_propagator_1():
    k = [-0.5, -0.2]

    def odefn(t, x, p, aux):
        return np.array([k[0]*x[0],k[1]*x[1]])

    y0 = np.array([10, -50])
    q0 = np.array([])
    tspan = np.array([0, 1.0])
    prop = Propagator()
    solout = prop(odefn, None, tspan, y0, q0, [], {})
    t1 = solout.t
    x1 = solout.y.T
    x1_expected = np.array([y*np.exp(k_*t1) for (y, k_) in zip(y0, k)])
    assert (x1 - x1_expected < 1e-5).all()

def test_propagator_2():
    def odefun(t, x, p, aux):
        return (-x[1], x[0])

    def quadfun(t, x, p, aux):
        return x[0]

    y0 = np.array([1, 0])
    q0 = np.array([0])
    tspan = np.array([0, pi / 2])
    prop = Propagator()
    solout = prop(odefun, quadfun, tspan, y0, q0, [], [])

    assert solout.y.T[0,-1] < tol
    assert solout.y.T[1,-1] - 1 < tol
    assert ((solout.y[:,1] - solout.q[:,0]) < tol).all()


def test_Trajectory():
    t = np.array([0,1,2,3])
    y1 = t ** 2

    gam = Trajectory(t, y1)
    y, q, u = gam(0.5)
    assert y == 0.5
    assert q == None
    assert u == None

    y, q, u = gam(0.25)
    assert y == 0.25
    assert q == None
    assert u == None

    gam.set_interpolate_function('cubic')

    y, q, u = gam(0.25)
    assert y - 0.0625 < 1e-4
    assert q == None
    assert u == None

    t, y, q, u = gam[0]
    assert t == 0
    assert y == 0
    assert q == None
    assert u == None


def test_integrate_quads():
    # Test a 1-dim x and 1-dim q
    t = np.linspace(0, 10, 100)
    y1 = np.sin(t)
    def quadfun(t, y):
        return y[0]

    gam = Trajectory(t, y1)
    assert integrate_quads(quadfun, np.array([0, 1 * np.pi]) + 0, gam) - 2 < 1e-3
    assert integrate_quads(quadfun, np.array([0, 2 * np.pi]) + 0, gam) < 1e-3
    assert integrate_quads(quadfun, np.array([0, 2 * np.pi]) + 1, gam) < 1e-3
    assert integrate_quads(quadfun, np.array([0, 2 * np.pi]) + 2, gam) < 1e-3
    assert integrate_quads(quadfun, np.array([0, 2 * np.pi]) + 3, gam) < 1e-3

    # Test a 2-dim x and 1-dim q
    t = np.linspace(0, 10, 100)
    y1 = np.sin(t)
    y2 = np.cos(t)

    def quadfun(t, y):
        return (y[0]*y[1])

    gam = Trajectory(t, np.vstack((y1, y2)).T)
    assert integrate_quads(quadfun, np.array([0, 1 * np.pi]) + 0, gam) - np.pi/2 < 1e-3
    assert integrate_quads(quadfun, np.array([0, 2 * np.pi]) + 0, gam) - np.pi < 1e-3

    # Test a 1-dim x and 2-dim q
    t = np.linspace(0, 10, 100)
    y1 = np.sin(t)

    def quadfun(t, y):
        return (y[0], y[0]**2)

    gam = Trajectory(t, y1)
    assert integrate_quads(quadfun, np.array([0, 1 * np.pi]) + 0, gam)[0] - 2 < 1e-3
    assert integrate_quads(quadfun, np.array([0, 1 * np.pi]) + 0, gam)[1] - np.pi/2 < 1e-3

    # Test a 2-dim x and 2-dim q
    t = np.linspace(0, 10, 100)
    y1 = np.sin(t)
    y2 = np.cos(t)

    def quadfun(t, y):
        return (y[0], y[1])

    gam = Trajectory(t, np.vstack((y1,y2)).T)
    assert len(integrate_quads(quadfun, np.array([0, 2 * np.pi]) + 2, gam)) == 2
    assert integrate_quads(quadfun, np.array([0, 2 * np.pi]) + 2, gam)[0] < 1e-3
    assert integrate_quads(quadfun, np.array([0, 2 * np.pi]) + 2, gam)[1] < 1e-3
    assert integrate_quads(quadfun, np.array([0, 1 * np.pi]) + 0, gam)[0] - 2 < 1e-3
    assert integrate_quads(quadfun, np.array([0, 1 * np.pi]) + 0, gam)[1] < 1e-3



import os
import inspect
import warnings
import copy
import logging

from beluga.codegen.codegen import *

import dill
import numpy as np
import collections as cl

from beluga import problem, helpers
from beluga.bvpsol import algorithms, Solution
from .utils import tic, toc
from .utils.keyboard import keyboard
# from beluga.optimlib import ocp_to_bvp

config = dict(logfile='beluga.log',
              default_bvp_solver='Shooting',
              output_file='data.dill')

BVP = cl.namedtuple('BVP', 'deriv_func bc_func compute_control')
def bvp_algorithm(algo, **kwargs):
    """
    Helper method to load algorithm by name
    """
    # Load algorithm from the package
    for name, obj in inspect.getmembers(algorithms):
        if inspect.isclass(obj):
            if name.lower() == algo.lower():
                return obj(**kwargs)
    else:
        # Raise exception if the loop completes without finding an algorithm
        # by the given name
        raise ValueError('Algorithm '+algo+' not found')


def guess_generator(*args, **kwargs):
    """Helper for creating Initial guess generator."""
    guess_gen = problem.GuessGenerator()
    guess_gen.setup(*args,**kwargs)
    return guess_gen


def setup_beluga(logging_level=logging.INFO, display_level=logging.INFO, output_file=None):
    """Performs initial configuration on beluga."""

    # Get reference to the input file module
    frm = inspect.stack()[1]
    input_module = (inspect.getmodule(frm[0]))

    # Get information about input file
    info = inspect.getframeinfo(frm[0])

    # Suppress warnings
    warnings.filterwarnings("ignore")

    # Initialize logging system
    helpers.init_logging(logging_level,display_level, config['logfile'])

    # Set the output file name
    if output_file is not None:
        config['output_file'] = output_file

    # solve(problem)
    return


def solve(ocp, method, bvp_algorithm, steps, guess_generator, output_file='data.dill'):
    """
    Solves the OCP using specified method
    """

    # Initialize necessary conditions of optimality object
    # print("Computing the necessary conditions of optimality")
    logging.info("Computing the necessary conditions of optimality")
    from beluga.optimlib import methods

    wf = methods[method]
    # ocp_ws = ocp_to_bvp(ocp, guess_generator)
    ocp_ws = wf({'problem': ocp, 'guess': guess_generator})

    solinit = Solution()

    solinit.aux['const'] = dict((str(const.name),float(const.value))
                                for const in ocp_ws['constants'])
    solinit.aux['parameters'] = ocp_ws['problem_data']['parameter_list']

    # For path constraints
    solinit.aux['constraint'] = cl.defaultdict(float)
    solinit.aux['constraints'] = dict((s['name'], {'unit':str(s['unit']),
                                                   'expr':str(s['expr']),
                                                   'direction': s['direction'],
                                                   'arc_type': i,
                                                   'pi_list':[str(_) for _ in s['pi_list']]})
                                      for i, s in enumerate(ocp_ws['problem_data']['s_list'],1))
    solinit.aux['arc_seq'] = (0,)
    solinit.aux['pi_seq'] = (None,)
    bvp_fn, bvp = bvp_algorithm.preprocess(ocp_ws['problem_data'])
    # The initial guess is automatically stored in the bvp object
    # solinit is just a reference to it
    solinit = ocp_ws['guess'].generate(bvp_fn, solinit)

    # # includes costates
    state_names = ocp_ws['problem_data']['state_list']

    initial_states = solinit.y[:,0] # First column
    terminal_states = solinit.y[:,-1] # Last column
    initial_bc = dict(zip(state_names,initial_states))
    terminal_bc = dict(zip(state_names,terminal_states))

    solinit.aux['initial'] = initial_bc
    solinit.aux['terminal'] = terminal_bc
    tic()
    # TODO: Start from specific step for restart capability
    # TODO: Make class to store result from continuation set?
    out = dict()

    out['problem_data'] = ocp_ws['problem_data']

    ocp._scaling.initialize(ocp_ws)
    ocp_ws['scaling'] = ocp._scaling

    out['solution'] = run_continuation_set(ocp_ws, bvp_algorithm, steps, bvp_fn, solinit, bvp)
    total_time = toc()

    # tic()
    # out['solution'] = run_continuation_set(ocp_ws, bvp_algorithm, steps, bvp_fn, solinit2)
    # total_time = toc()

    logging.info('Continuation process completed in %0.4f seconds.\n' % total_time)
    bvp_algorithm.close()

    # Final time is appended as a parameter, so scale the output x variables to show the correct time
    for continuation_set in out['solution']:
        for sol in continuation_set:
            tf_ind = [i for i, s in enumerate(out['problem_data']['parameter_list']) if s is 'tf'][0]
            tf = sol.parameters[tf_ind]
            sol.t = sol.t*tf

    # Save data
    # del out['problem_data']['s_list']
    del out['problem_data']['states']
    del out['problem_data']['costates']

    qvars = out['problem_data']['quantity_vars']
    qvars = {str(k):str(v) for k,v in qvars.items()}
    out['problem_data']['quantity_vars'] = qvars
    with open(output_file, 'wb') as outfile:
        dill.settings['recurse'] = True
        dill.dump(out, outfile) # Dill Beluga object only

    return out['solution'][-1][-1]


def run_continuation_set(ocp_ws, bvp_algo, steps, bvp_fn, solinit, bvp):
    # Loop through all the continuation steps
    solution_set = []
    # Initialize scaling
    s = ocp_ws['scaling']
    problem_data = ocp_ws['problem_data']
    try:
        sol_guess = solinit
        sol = None
        for step_idx, step in enumerate(steps):
            logging.info('\nRunning Continuation Step #'+str(step_idx+1)+' : ')
            solution_set.append([])
            # Assign solution from last continuation set
            step.reset()
            step.init(sol_guess, problem_data)

            for aux in step:  # Continuation step returns 'aux' dictionary
                sol_guess.aux = aux

                logging.info('Starting iteration '+str(step.ctr)+'/'+str(step.num_cases()))
                tic()

                s.compute_scaling(sol_guess)
                sol_guess = s.scale(sol_guess)

                sol = bvp_algo.solve(bvp.deriv_func, None, bvp.bc_func, sol_guess)
                step.last_sol.converged = sol.converged
                sol = s.unscale(sol)

                if sol.converged:
                    # Post-processing phase

                    # Compute control history
                    # Required for plotting to work with control variables
                    sol.ctrl_expr = problem_data['control_options']
                    sol.ctrl_vars = problem_data['control_list']

                    #TODO: Make control computation more efficient
                    # for i in range(len(sol.x)):
                    #     _u = bvp.control_func(sol.x[i],sol.y[:,i],sol.parameters,sol.aux)
                    #     sol.u[:,i] = _u

                    ## DAE mode
                    # sol.u = sol.y[problem_data['num_states']:,:]
                    # Non-DAE:
                    f = lambda _t, _X: bvp_fn.compute_control(_t,_X,sol.parameters,sol.aux)
                    sol.u = np.array(list(map(f, sol.t, list(sol.y.T)))).T

                    # Copy solution object for storage and reuse `sol` in next
                    # iteration
                    solution_set[step_idx].append(copy.deepcopy(sol))
                    sol_guess = copy.deepcopy(sol)
                    elapsed_time = toc()
                    logging.info('Iteration %d/%d converged in %0.4f seconds\n' % (step.ctr, step.num_cases(), elapsed_time))
                else:
                    elapsed_time = toc()
                    logging.info('Iteration %d/%d failed to converge!\n' % (step.ctr, step.num_cases()))
    except Exception as e:
        import traceback
        traceback.print_exc()
        logging.error('Exception : ' + str(e))
        logging.error('Stopping')

    return solution_set



import abc
class BaseRenderer(object):
    # Define class as abstract class
    __metaclass__ = abc.ABCMeta

    # Define common interface for renderer classes
    @abc.abstractmethod
    def create_figure(self):
        """
        Creates a new figure and returns a handle
        """

    @abc.abstractmethod
    def close_figure(self,figure):
        """
        Closes a specified figure
        """

    @abc.abstractmethod
    def show_figure(self,figure):
        """
        Shows a specified figure
        """

    @abc.abstractmethod
    def show_all(self):
        """
        Shows all rendered figures
        """

    @abc.abstractmethod
    def render_plot(self,figure,plot):
        """
        Adds a line plot using the given data to the specified figure
        """

    @abc.abstractmethod
    def render_subplot(self,figure,plot):
        """
        Adds a subplot to the specified figure
        """

    # Configuration
    # Movie making interface
    # Method for saving plot to file (PDF/SVG/JPEG/PNG/HTML5)
